/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.database.backend.implementation.MySQL;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import org.caosdb.server.database.access.Access;
import org.caosdb.server.database.backend.interfaces.GetFileEntityByPathImpl;
import org.caosdb.server.database.exceptions.TransactionException;
import org.caosdb.server.database.proto.SparseEntity;
import org.caosdb.server.filesystem.Path;

/**
 * Implements {@link GetFileEntityByPathImpl} for a MySQL/MariaDB back-end.
 *
 * @author Timm Fitschen <t.fitschen@indiscale.com>
 */
public class MySQLGetFileEntityByPath extends MySQLTransaction implements GetFileEntityByPathImpl {

  public MySQLGetFileEntityByPath(final Access access) {
    super(access);
  }

  public static final String STMT_GET_FSO_BY_PATH_LIKE = "call retrieveFSODescriptorByPath(?)";
  public static final String STMT_GET_ACL = "call retrieveEntityACL(?)";

  @Override
  public SparseEntity execute(final Path path, final boolean like) throws TransactionException {
    try {
      final SparseEntity result = retrieveFileDescriptor(path.toString(), like);
      if (result == null) {
        return null;
      }
      return retrieveEntityACL(result);
    } catch (final SQLException e) {
      throw new TransactionException(e);
    } catch (final ConnectionException e) {
      throw new TransactionException(e);
    }
  }

  private SparseEntity retrieveFileDescriptor(final String path, final boolean like)
      throws SQLException, ConnectionException {
    final PreparedStatement stmt;
    stmt = prepareStatement(STMT_GET_FSO_BY_PATH_LIKE);
    if (like) {
      stmt.setString(1, path + "%");
    } else {
      stmt.setString(1, path);
    }

    try (ResultSet rs = stmt.executeQuery()) {
      if (rs.next()) {
        final SparseEntity ret = new SparseEntity();
        ret.id = rs.getString("FileId");
        DatabaseUtils.parseFSODescriptorFields(rs, ret);
        return ret;
      } else {
        return null;
      }
    }
  }

  private SparseEntity retrieveEntityACL(final SparseEntity entity)
      throws SQLException, ConnectionException {
    final PreparedStatement stmt = prepareStatement(STMT_GET_ACL);
    stmt.setString(1, entity.id);
    try (ResultSet rs = stmt.executeQuery()) {
      rs.next();
      entity.acl = DatabaseUtils.bytes2UTF8(rs.getBytes("ACL"));
    }
    return entity;
  }
}
