package org.caosdb.server.database.backend.implementation.MySQL;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.LinkedList;
import java.util.List;
import org.caosdb.server.database.access.Access;
import org.caosdb.server.database.backend.interfaces.ListFilesImpl;
import org.caosdb.server.database.exceptions.TransactionException;
import org.caosdb.server.database.proto.SparseEntity;
import org.caosdb.server.filesystem.FSODescriptor;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.filesystem.Path;
import org.caosdb.server.filesystem.VirtualFSODescriptorInterface;

class LiveFSODescriptor extends FSODescriptor {

  private PreparedStatement pstmt;

  public LiveFSODescriptor(SparseEntity spe, PreparedStatement pstmt) {
    super(spe);
    this.pstmt = pstmt;
  }

  @Override
  public List<VirtualFSODescriptorInterface> listChildren() {
    List<VirtualFSODescriptorInterface> children = super.listChildren();
    if (getObjectType() == ObjectType.DIRECTORY && children == null) {
      // resolve here and now
      children = new LinkedList<VirtualFSODescriptorInterface>();
      try {
        pstmt.setString(1, getEntityId().toString());
        try (ResultSet rs = pstmt.executeQuery()) {
          while (rs.next()) {
            final SparseEntity entity = new SparseEntity();
            entity.id = rs.getString("FileId");
            DatabaseUtils.parseFSODescriptorFields(rs, entity);
            children.add(new LiveFSODescriptor(entity, pstmt));
          }
        }

      } catch (SQLException e) {
        throw new TransactionException(e);
      }
    }
    return children;
  }
}

public class MySQLListFiles extends MySQLTransaction implements ListFilesImpl {

  public MySQLListFiles(Access access) {
    super(access);
  }

  public static final String LIST_CHILDREN_STMT =
      "SELECT 'SHA-512' AS FileHashAlgo, file_id AS FileId, "
          + "parent_directory AS FileParentID, path AS FilePath, size AS FileSize, hex(hash) AS FileHash, checked_timestamp AS FileHashChecked, mimetype AS FileMimeType, file_storage_id AS FileStorageID, file_key AS FileKey FROM files WHERE parent_directory = ?";
  public static final String LIST_FILES_STMT =
      "SELECT 'SHA-512' AS FileHashAlgo, file_id AS FileId, "
          + "parent_directory AS FileParentID, path AS FilePath, size AS FileSize, hex(hash) AS FileHash, checked_timestamp AS FileHashChecked, mimetype AS FileMimeType, file_storage_id AS FileStorageID, file_key AS FileKey FROM files";
  public static final String FILE_STORAGE_CLAUSE = " file_storage_id = ?";
  public static final String PATH_LIKE_CLAUSE = " path LIKE ?";
  public static final String LAST_CHECKED_BEFORE_CLAUSE = " checked_timestamp < ?";
  public static final String WHERE_TOKEN = " WHERE";
  public static final String AND_TOKEN = " AND";
  public static final String LIMIT = " LIMIT ";

  @Override
  public List<VirtualFSODescriptorInterface> execute(
      String fileStorageId, Path path, Long lastChecked, Integer limit) {

    final List<VirtualFSODescriptorInterface> result = new LinkedList<>();
    try {
      PreparedStatement stmt = prepareStatement(fileStorageId, path, lastChecked, limit);
      PreparedStatement listChildrenStmt = prepareStatement(LIST_CHILDREN_STMT);
      try (ResultSet rs = stmt.executeQuery()) {
        while (rs.next()) {
          final SparseEntity entity = new SparseEntity();
          entity.id = rs.getString("FileId");
          DatabaseUtils.parseFSODescriptorFields(rs, entity);
          result.add(new LiveFSODescriptor(entity, listChildrenStmt));
        }
      }
    } catch (SQLException e) {
      throw new TransactionException(e);
    } catch (ConnectionException e) {
      throw new TransactionException(e);
    }

    return result;
  }

  private PreparedStatement prepareStatement(
      String fileStorageId, Path path, Long lastChecked, Integer limit)
      throws SQLException, ConnectionException {
    String stmtStr = LIST_FILES_STMT;
    String token = WHERE_TOKEN;

    if (use(fileStorageId)) {
      stmtStr += token + FILE_STORAGE_CLAUSE;
      token = AND_TOKEN;
    }

    if (use(path)) {
      stmtStr += token + PATH_LIKE_CLAUSE;
      token = AND_TOKEN;
    }

    if (use(lastChecked)) {
      stmtStr += token + LAST_CHECKED_BEFORE_CLAUSE;
    }

    if (limit != null) {
      stmtStr += LIMIT + Integer.toString(limit);
    }

    PreparedStatement result = prepareStatement(stmtStr);

    // be sure to keep these in the exact same order as above and use the exact same conditions.
    int parameter = 1;
    if (use(fileStorageId)) {
      result.setString(parameter, fileStorageId);
      parameter++;
    }

    if (use(path)) {
      result.setString(parameter, path.toString());
      parameter++;
    }

    if (use(lastChecked)) {
      result.setLong(parameter, lastChecked);
    }

    return result;
  }

  private boolean use(Long lastChecked) {
    return lastChecked != null && lastChecked > 0;
  }

  private boolean use(Path path) {
    return path != null && !path.isEmpty();
  }

  private boolean use(String fileStorageId) {
    return fileStorageId != null && !fileStorageId.isBlank();
  }
}
