/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.database.backend.implementation.MySQL;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLIntegrityConstraintViolationException;
import java.sql.Types;
import org.caosdb.server.database.access.Access;
import org.caosdb.server.database.backend.interfaces.UpdateSparseEntityImpl;
import org.caosdb.server.database.exceptions.IntegrityException;
import org.caosdb.server.database.exceptions.TransactionException;
import org.caosdb.server.database.proto.SparseEntity;
import org.caosdb.server.filesystem.FileSystem.ObjectType;

public class MySQLUpdateSparseEntity extends MySQLTransaction implements UpdateSparseEntityImpl {

  public MySQLUpdateSparseEntity(final Access access) {
    super(access);
  }

  public static final String STMT_UPDATE_ENTITY = "call updateEntity(?,?,?,?,?,?,?)";

  public static final String STMT_UPDATE_FILE_PROPS = "call setFileProperties(?,?,?,?,?,?,?,?,?,?)";

  @Override
  public void execute(final SparseEntity spe) throws TransactionException {
    try {
      // file properties;
      final PreparedStatement updateFilePropsStmt = prepareStatement(STMT_UPDATE_FILE_PROPS);
      updateFilePropsStmt.setString(1, spe.id);
      if (spe.filePath != null) {
        updateFilePropsStmt.setString(2, spe.filePath.toString()); // FilePath
        updateFilePropsStmt.setLong(3, spe.fileSize); // FileSize
        if (spe.fileHash != null) {
          updateFilePropsStmt.setString(4, spe.fileHash.value); // FileHash
          updateFilePropsStmt.setLong(5, spe.fileHash.datetime); // FileChecked
          updateFilePropsStmt.setString(10, spe.fileHash.algorithm); // FileHashAlgo
        } else {
          updateFilePropsStmt.setNull(4, Types.VARCHAR); // FileHash
          updateFilePropsStmt.setLong(5, 0L); // FileChecked
          updateFilePropsStmt.setNull(10, Types.VARCHAR); // FileHashAlgo
        }
        if (spe.fileType == ObjectType.DIRECTORY) {
          updateFilePropsStmt.setString(6, "inode/directory");
        } else if (spe.fileType == ObjectType.LINK) {
          updateFilePropsStmt.setString(6, "inode/link");
        } else if (spe.fileMimeType != null) {
          updateFilePropsStmt.setString(6, spe.fileMimeType); // FileMimetype
        } else {
          updateFilePropsStmt.setNull(6, Types.VARBINARY); // FileMimetype
        }
        updateFilePropsStmt.setString(7, spe.fileStorageId); // FileStorageId
        updateFilePropsStmt.setString(8, spe.fileKey); // FileKey
        if (spe.fileParentId != null) {
          updateFilePropsStmt.setString(9, spe.fileParentId);
        } else {
          updateFilePropsStmt.setNull(9, Types.VARCHAR); // FileParentDirectory
        }
      } else {
        // move file to archive
        updateFilePropsStmt.setNull(2, Types.VARCHAR); // FilePath
        updateFilePropsStmt.setNull(3, Types.BIGINT); // FileSize
        updateFilePropsStmt.setNull(4, Types.VARCHAR); // FileHash
        updateFilePropsStmt.setNull(5, Types.BIGINT); // FileChecked
        updateFilePropsStmt.setNull(6, Types.VARBINARY); // FileMimetype
        updateFilePropsStmt.setNull(7, Types.VARBINARY); // FileStorageId
        updateFilePropsStmt.setNull(8, Types.VARBINARY); // FileKey
        updateFilePropsStmt.setNull(9, Types.INTEGER); // FileParentDirectory
        updateFilePropsStmt.setNull(10, Types.VARCHAR); // FileHashAlgo
      }
      updateFilePropsStmt.execute();

      // very sparse entity
      final PreparedStatement updateEntityStmt = prepareStatement(STMT_UPDATE_ENTITY);
      updateEntityStmt.setString(1, spe.id);
      updateEntityStmt.setString(2, spe.name);
      updateEntityStmt.setString(3, spe.description);
      updateEntityStmt.setString(4, spe.role);
      if (spe.datatype_id != null) {
        updateEntityStmt.setString(5, spe.datatype_id);
      } else {
        updateEntityStmt.setNull(5, Types.VARCHAR);
      }
      updateEntityStmt.setString(6, spe.collection);
      updateEntityStmt.setString(7, spe.acl);
      final ResultSet rs = updateEntityStmt.executeQuery();

      if (rs.next()) {
        spe.versionId = DatabaseUtils.bytes2UTF8(rs.getBytes("Version"));
      }

    } catch (final SQLIntegrityConstraintViolationException e) {
      throw new IntegrityException(e);
    } catch (final SQLException e) {
      throw new TransactionException(e);
    } catch (final ConnectionException e) {
      throw new TransactionException(e);
    }
  }
}
