package org.caosdb.server.database.backend.transaction;

import org.caosdb.server.database.BackendTransaction;
import org.caosdb.server.database.exceptions.EntityDoesNotExistException;
import org.caosdb.server.entity.EntityID;
import org.caosdb.server.entity.EntityInterface;
import org.caosdb.server.entity.Message;
import org.caosdb.server.entity.container.TransactionContainer;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.filesystem.Path;
import org.caosdb.server.filesystem.VirtualFSODescriptorInterface;
import org.caosdb.server.utils.ServerMessages;

/**
 * Checks if a target path for a new file or a file which is to be moved is ok.
 *
 * <p>That means, there must not exist any other file entity or directory entity with this exact
 * path.
 *
 * @author Timm Fitschen (t.fitschen@indiscale.com)
 */
public class CheckTargetPath extends BackendTransaction {

  private final EntityInterface entity;
  private TransactionContainer container;

  public CheckTargetPath(final EntityInterface entity, TransactionContainer container) {
    this.container = container;
    this.entity = entity;
  }

  @Override
  protected void execute() {
    try {
      checkTarget(entity, container);
    } catch (final Message e) {
      entity.addError(e);
    }
  }

  public void checkTarget(final EntityInterface entity, TransactionContainer container)
      throws Message {

    checkNastyPath(entity.getFSODescriptor().getPath());
    checkOwner(entity);
    checkParentExists(entity, container);
    checkLinkTarget(entity, container);
  }

  private void checkLinkTarget(EntityInterface entity, TransactionContainer container) {
    if (entity.getFSODescriptor().getObjectType() != ObjectType.LINK) {
      return;
    }

    EntityID linkTarget = entity.getFSODescriptor().getLinkTarget();

    EntityInterface linkTargetEntity = container.getEntityById(linkTarget);
    if (linkTargetEntity == null) {
      RetrieveSparseEntity t = new RetrieveSparseEntity(linkTarget, null);
      linkTargetEntity = execute(t).getEntity();
    }

    if (linkTargetEntity.getRole().hasFSODescriptor()) {
      linkTarget.link(linkTargetEntity.getId());
      entity.getFSODescriptor().setLinkTargetPath(linkTargetEntity.getFSODescriptor().getPath());

    } else {
      // TODO(tf) throw something
    }
  }

  private void checkParentExists(EntityInterface entity, TransactionContainer container)
      throws Message {
    Path path = entity.getFSODescriptor().getPath().getParent();
    if (path == null) {
      // root is parent
      return;
    }
    for (EntityInterface e : container) {
      if (e.getFSODescriptor() != null && e.getFSODescriptor().getPath() != null) {
        if (e.getFSODescriptor().getPath().equals(path)) {
          // found the parent in the current container
          return;
        }
      }
    }

    GetFileEntityByPath t = new GetFileEntityByPath(path, false);
    try {
      execute(t);
    } catch (final EntityDoesNotExistException e) {
      throw ServerMessages.TARGET_PARENT_DIRECTORY_DOES_NOT_EXIST;
    }
  }

  private void checkOwner(EntityInterface entity) throws Message {
    final VirtualFSODescriptorInterface fd = entity.getFSODescriptor();
    final Path path = fd.getPath();
    final GetFileEntityByPath t = new GetFileEntityByPath(path, false);
    try {
      execute(t);
    } catch (final EntityDoesNotExistException e) {
      // No entity owns this path
      return;
    }

    if (t.getEntity().getId().equals(entity.getId())) {
      // this entity already owns this path.
      return;
    } else {
      throw ServerMessages.TARGET_PATH_EXISTS;
    }
  }

  private void checkNastyPath(Path path) throws Message {
    for (final String segment : path.getSegments()) {
      if (segment.matches("^\\.{1,3}$")) {
        throw ServerMessages.TARGET_PATH_NOT_ALLOWED;
      }
    }
  }
}
