/*
 * ** header v3.0 This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics, Max-Planck-Institute for Dynamics and
 * Self-Organization Göttingen Copyright (C) 2019 IndiScale GmbH Copyright (C) 2019 Timm Fitschen
 * (t.fitschen@indiscale.com)
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the
 * GNU Affero General Public License as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
 * even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License along with this program.
 * If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.database.backend.transaction;

import java.util.LinkedList;
import java.util.List;
import org.apache.commons.jcs.access.behavior.ICacheAccess;
import org.caosdb.server.database.CacheableBackendTransaction;
import org.caosdb.server.database.backend.interfaces.GetFileEntityByPathImpl;
import org.caosdb.server.database.backend.interfaces.GetFilesInDirectoryImpl;
import org.caosdb.server.database.exceptions.EntityDoesNotExistException;
import org.caosdb.server.database.exceptions.TransactionException;
import org.caosdb.server.database.proto.SparseEntity;
import org.caosdb.server.entity.EntityID;
import org.caosdb.server.entity.EntityInterface;
import org.caosdb.server.entity.RetrieveEntity;
import org.caosdb.server.filesystem.FSODescriptor;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.filesystem.Hash;
import org.caosdb.server.filesystem.Path;
import org.caosdb.server.filesystem.VirtualFSODescriptorInterface;
import org.caosdb.server.permissions.EntityACL;

/**
 * Retrieve a file entity by the path of the file.
 *
 * @author Timm Fitschen (t.fitschen@indiscale.com)
 */
public class GetFileEntityByPath extends CacheableBackendTransaction<String, SparseEntity> {

  private static final ICacheAccess<String, SparseEntity> cache = null;
  // TODO: caching is currently disabled, because the cache objects don't
  // invalidate correctly which causes wrong results.
  //      Cache.getCache("BACKEND_SparseFileRecordsByPath");
  private final Path path;
  private EntityInterface entity;
  private final boolean like;

  public GetFileEntityByPath(final Path path) {
    this(path, false);
  }

  public GetFileEntityByPath(final Path path, final boolean like) {
    super(cache);
    this.path = path;
    this.like = like;
  }

  /**
   * TODO removeCached doesn't work, if a file was moved! This is why the cache doesn't work.
   *
   * @param path
   */
  public static void removeCached(final Path path) {
    if (path != null && cache != null) {
      cache.remove("1" + path.toString());
      cache.remove("0" + path.toString());
    }
  }

  @Override
  protected String getKey() {
    return (this.like ? "1" : "0") + this.path.toString();
  }

  @Override
  protected void process(final SparseEntity t) throws TransactionException {
    this.entity = parseEntity(t);
  }

  private EntityInterface parseEntity(final SparseEntity spe) {
    final EntityInterface entity = new RetrieveEntity(new EntityID(spe.id));
    entity.setEntityACL(EntityACL.deserialize(spe.acl));

    final VirtualFSODescriptorInterface fd = parseFileDescriptor(spe);

    if (fd.getObjectType() == ObjectType.DIRECTORY) {
      fd.setChildren(getChildren(fd));
    }

    entity.setFSODescriptor(fd);
    return entity;
  }

  private VirtualFSODescriptorInterface parseFileDescriptor(final SparseEntity spe) {
    return new FSODescriptor(spe);
  }

  @Override
  public SparseEntity executeNoCache() throws TransactionException {
    final GetFileEntityByPathImpl t = getImplementation(GetFileEntityByPathImpl.class);
    final SparseEntity result = t.execute(path, like);
    if (result == null) {
      throw new EntityDoesNotExistException();
    }
    return result;
  }

  /**
   * Return a list of all subdirectories and files in a directory.
   *
   * @param directory
   * @return
   */
  private List<VirtualFSODescriptorInterface> getChildren(
      final VirtualFSODescriptorInterface directory) {
    // TODO move to separate transaction and remove listFiles() from FileDescriptorInterface
    final GetFilesInDirectoryImpl t = getImplementation(GetFilesInDirectoryImpl.class);
    final List<SparseEntity> spes = t.execute(directory.getEntityId());
    final List<VirtualFSODescriptorInterface> result = new LinkedList<>();
    for (final SparseEntity spe : spes) {
      final VirtualFSODescriptorInterface fd = parseFileDescriptor(spe);
      result.add(fd);
    }

    return result;
  }

  public EntityInterface getEntity() {
    return this.entity;
  }

  public Long getSize() {
    return this.entity.getFSODescriptor().getSize();
  }

  public Hash getHash() {
    return this.entity.getFSODescriptor().getHash();
  }

  public EntityID getId() {
    return this.entity.getId();
  }
}
