package org.caosdb.server.filesystem;

import java.util.List;
import org.caosdb.server.database.proto.SparseEntity;
import org.caosdb.server.entity.EntityID;
import org.caosdb.server.entity.EntityInterface;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.utils.EntityStatus;

/**
 * Represents the objects of the virtual file system. These objects represent the world as it SHOULD
 * BE - The file's hash(es), size and mime type, when it is intact, the directories children when no
 * file has gone missing, the links target etc. This object does not have direct access to the files
 * blob.
 *
 * <p>This is mainly a data-class which holds information and is being attached to an entity (see
 * {@link EntityInterface}).
 *
 * <p>An instance of this class is the twin of an instance of the {@link RealFSODescriptorInterface}
 * which represents the file as it ACTUALLY IS. The twins are associated via their matching {@link
 * #fileStorageId} and {@link #key} which are unique identifiers of any FSO (FileSystemObject), be
 * it a file, a directory or a symbolic link.
 *
 * <p>N.B. For each {@link * VirtualFSODescriptorInterface} there is a matching {@link
 * RealFSODescriptorInterface}. This reverse is not necessarily true.
 *
 * @author Timm Fitschen <t.fitschen@indiscale.com>
 */
public class FSODescriptor implements VirtualFSODescriptorInterface {

  protected String fileStorageId;
  protected String key;
  protected EntityStatus status = null;
  protected Hash hash = null;
  protected Path path = null;
  protected Path linkTargetPath = null;
  protected Long size = null;
  protected String tmpIdentifier = null;
  protected EntityID entityId = new EntityID();
  protected String mimeType;
  protected FileSystem.ObjectType type;
  protected EntityID parentDirectory = null;
  protected List<VirtualFSODescriptorInterface> children;
  private RealFSODescriptorInterface twin;
  private EntityID linkTarget = null;

  public FSODescriptor(
      final String fileStorageId,
      final String key,
      final Hash hash,
      final Path path,
      final Long size,
      final String tmpIdentifier) {
    this.fileStorageId = fileStorageId;
    this.key = key;
    this.hash = hash;
    this.path = path;
    this.size = size;
    this.tmpIdentifier = tmpIdentifier;
  }

  public FSODescriptor(
      final String fileStorageId,
      final String key,
      final Hash hash,
      final Path path,
      final Long size) {
    this(fileStorageId, key, hash, path, size, null);
  }

  public FSODescriptor(final Path path) {
    this.setPath(path);
  }

  public FSODescriptor(final SparseEntity spe) {
    this.hash = spe.fileHash;
    this.size = spe.fileSize;
    this.path = spe.filePath;
    this.entityId = new EntityID(spe.id);
    this.fileStorageId = spe.fileStorageId;
    this.key = spe.fileKey;
    this.type = spe.fileType;
    this.mimeType = spe.fileMimeType;
    this.parentDirectory = new EntityID(spe.fileParentId);
    this.linkTarget = new EntityID(spe.linkTarget);
  }

  public FSODescriptor(final Path path, final Long size) {
    this.path = path;
    this.size = size;
  }

  @Override
  public EntityStatus getEntityStatus() {
    return status;
  }

  @Override
  public void setEntityStatus(final EntityStatus status) {
    this.status = status;
  }

  @Override
  public boolean hasTmpIdentifier() {
    return tmpIdentifier != null;
  }

  @Override
  public void setTmpIdentifier(final String tmpIdentifier) {
    this.tmpIdentifier = tmpIdentifier;
  }

  @Override
  public String getTmpIdentifier() {
    return tmpIdentifier;
  }

  @Override
  public void setSize(final Long size) {
    this.size = size;
  }

  @Override
  public void setPath(final Path path) {
    this.path = path;
  }

  @Override
  public boolean hasPath() {
    return path != null;
  }

  @Override
  public boolean hasSize() {
    return size != null;
  }

  @Override
  public boolean hasHash() {
    return hash != null;
  }

  @Override
  public Path getPath() {
    return path;
  }

  @Override
  public Long getSize() {
    return size;
  }

  @Override
  public String getKey() {
    return key;
  }

  @Override
  public String getFileStorageId() {
    return fileStorageId;
  }

  @Override
  public EntityID getEntityId() {
    return entityId;
  }

  @Override
  public String getMimeType() {
    return mimeType;
  }

  @Override
  public String getName() {
    return getPath().getLastSegment();
  }

  @Override
  public List<VirtualFSODescriptorInterface> listChildren() {
    return children;
  }

  @Override
  public void setMimeType(final String mimeType) {
    this.mimeType = mimeType;
  }

  @Override
  public void setChildren(final List<VirtualFSODescriptorInterface> children) {
    this.children = children;
  }

  public static FSODescriptor createDir(String fileStorageId, final Path dirPath) {
    final FSODescriptor ret = new FSODescriptor(dirPath);
    ret.key = dirPath.toString();
    ret.fileStorageId = fileStorageId;
    ret.type = ObjectType.DIRECTORY;
    return ret;
  }

  @Override
  public Hash getHash() {
    return this.hash;
  }

  @Override
  public void setHash(final Hash hash) {
    this.hash = hash;
  }

  @Override
  public void updateHash(final Hash hash) {
    if (hash.datetime > this.hash.datetime) {
      // the other hash is more recent
      setHash(hash);
    }
  }

  @Override
  public EntityID getParentDirectoryID() {
    return parentDirectory;
  }

  @Override
  public void setParentDirectory(final EntityID id) {
    parentDirectory = id;
  }

  @Override
  public void setParentDirectory(final VirtualFSODescriptorInterface link) {
    parentDirectory = new EntityID();
    parentDirectory.link(link.getEntityId());
  }

  @Override
  public String toString() {
    return getFileStorageId() + ":" + getKey();
  }

  @Override
  public ObjectType getObjectType() {
    return type;
  }

  @Override
  public EntityID getLinkTarget() {
    return linkTarget;
  }

  @Override
  public RealFSODescriptorInterface getTwin() {
    return twin;
  }

  @Override
  public void setTwin(RealFSODescriptorInterface twin) {
    this.twin = twin;
  }

  @Override
  public void setKey(String key) {
    this.key = key;
  }

  @Override
  public Path getLinkTargetPath() {
    return linkTargetPath;
  }

  public void setLinkTarget(EntityID linkTarget) {
    this.linkTarget = linkTarget;
  }

  @Override
  public void setLinkTargetPath(Path linkTargetPath) {
    this.linkTargetPath = linkTargetPath;
  }

  public static FSODescriptor createLink(
      String fileStorageId, final Path path, final EntityID target) {
    FSODescriptor result = new FSODescriptor(path);
    result.fileStorageId = fileStorageId;
    result.linkTarget = target;
    result.type = ObjectType.LINK;
    return result;
  }
}
