package org.caosdb.server.filesystem;

import java.util.List;
import org.caosdb.server.filesystem.FileSystem.ObjectType;

/**
 * The FSODescriptorInterface is the unified interface to all things related to the storage of
 * files, the mapping of actual byte blobs and streams to file entities and the inherent virtual
 * directory structure of CaosDB's file system.
 *
 * <p>This is the kind of object which file entities carry around and which are used to access the
 * file system for reading, writing and consistency checks and other book-keeping
 *
 * @author Timm Fitschen (t.fitschen@indiscale.com)
 */
public interface FSODescriptorInterface extends Comparable<FSODescriptorInterface> {

  /**
   * An identifier of the file. The key is guaranteed to be unique across the file storage where
   * this file's blob is stored.
   *
   * @return the file key
   */
  public abstract String getKey();

  /**
   * The file storage id identities the back-end which stores the actual byte blob. A the same time
   * it serves as the name space for the {@link #getKey()}.
   *
   * @return the id of the file storage where this file's blob is stored.
   */
  public abstract String getFileStorageId();

  /** Return the hash checksum of the file blob or null. */
  public abstract Hash getHash();

  /** Return the byte size of the file blob or null. */
  public abstract Long getSize();

  /**
   * Return the name of this file system object.
   *
   * @return
   */
  public abstract String getName();

  /**
   * If this is a directory, return a list of all children, otherwise null.
   *
   * @return All files and directories within this directory or null.
   */
  public abstract List<? extends FSODescriptorInterface> listChildren();

  /** Return the mime type or null. */
  public abstract String getMimeType();

  public abstract ObjectType getObjectType();

  @Override
  default int compareTo(FSODescriptorInterface other) {
    int fileStorageCompared = other.getFileStorageId().compareTo(getFileStorageId());
    if (fileStorageCompared == 0) {
      return other.getKey().compareTo(getKey());
    }
    return fileStorageCompared;
  }
}
