package org.caosdb.server.filesystem;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

/**
 * The capabilities of a particular file storage.
 *
 * <p>E.g. A local file system should normally be readable and writable, while a remote file system
 * might even be non-readable.
 *
 * @author tf
 */
public class FileStorageCapabilities {

  public static final FileStorageCapabilities FULL = createFull();

  public static final FileStorageCapabilities READ_ONLY = createRO();

  /** This file storage can determine if a file still exists. */
  public boolean existence;

  /**
   * Files in this file storage can be read.
   *
   * <p>File storages without this capability do not provide a means to read or download a file via
   * CaosDB.
   */
  public boolean read;

  /*
   * Future:
   *
   * Files in this file storage are appendable.
   *
   * Clients can append a byte-blob to a file instead of replacing the file entirely.
   */
  // public boolean append = false;

  /*
   * Future:
   *
   * Files in this file storage are partially replacable, e.g. single cells in a csv file.
   *
   */
  // public boolean partialReplace = false;

  /** Files in this file storage can be replaced (entirely). */
  public boolean replace;

  /** Files in this file storage can be deleted. */
  public boolean delete;

  /** New files can be inserted into this file storage. */
  public boolean insert;

  /** The file storage can list all files */
  public boolean list;

  /** The file storage supports directory objects. */
  public boolean directory;

  /** The file storage can determine the size of a file. */
  public boolean getSize;

  /** The file storage supports these hash algorithms. */
  public Set<String> hashAlgorithms;

  private static FileStorageCapabilities createFull() {
    FileStorageCapabilities result = new FileStorageCapabilities();
    result.existence = true;
    result.read = true;
    result.insert = true;
    result.delete = true;
    result.replace = true;
    result.list = true;
    result.directory = true;
    result.getSize = true;
    String[] algos = {"SHA-512"};
    result.hashAlgorithms = new HashSet<>(Arrays.asList(algos));
    return result;
  }

  /** TODO rename */
  private static FileStorageCapabilities createRO() {
    FileStorageCapabilities result = createFull();
    result.insert = false;
    result.delete = false;
    result.replace = false;
    return result;
  }

  public boolean getHash(String algo) {
    return this.hashAlgorithms.contains(algo);
  }
}
