package org.caosdb.server.filesystem;

import org.caosdb.server.entity.Message;
import org.caosdb.server.utils.Undoable;

/**
 * FileStorageInterface is an interface for a descriptor of file system back-ends. It encapsulate
 * the various APIs of the actual back-end implementation.
 *
 * @author Timm Fitschen (t.fitschen@indiscale.com)
 */
public interface FileStorageInterface {

  /**
   * The file storage's id. This is a namespace of file system objects, which are identified by
   * their key, which must be unique for a file storage.
   *
   * @return the id.
   */
  public abstract String getId();

  /**
   * Return the capabilities of this file storage.
   *
   * <p>The capabilities are a set of supported operations. Some file storages might support
   * calculating SHA-512 hashes, while others do not. Some might support writing or reading file
   * blobs, others do not.
   *
   * @return {@link FileStorageCapabilities}
   */
  public abstract FileStorageCapabilities getCapabilities();

  /**
   * Return a file system object (see {@link RealFSODescriptorInterface}) which represents a single
   * file, directory or link that is stored in this file storage.
   *
   * @param fileKey - the unique identifier of the file system object to be resolved.
   * @return
   */
  public abstract RealFSODescriptorInterface resolve(String fileKey);

  public abstract Undoable move(
      RealFSODescriptorInterface file, VirtualFSODescriptorInterface target) throws Message;

  public abstract Undoable delete(String fileKey) throws Message;

  public abstract Long getSize(FSODescriptorInterface fso);

  public abstract Hash getHash(FSODescriptorInterface fso, String algorithm);

  /** Return true iff an object with this key exists. */
  public abstract boolean exists(String key);

  public abstract Iterable<? extends RealFSODescriptorInterface> list(String prefix);
}
