/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2022 IndiScale GmbH <info@indiscale.com>
 * Copyright (C) 2022 Timm Fitschen <t.fitschen@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */
package org.caosdb.server.filesystem;

import java.io.Serializable;

/**
 * Represents a hash (aka checksum) of a file entity.
 *
 * <p>Instances are either created by the {@link Hasher} class via the {@link #create(String, long,
 * String)} factory method.
 *
 * @author Timm Fitschen <t.fitschen@indiscale.com>
 */
public final class Hash implements Serializable {

  private static final long serialVersionUID = 70651761751684844L;

  /** The hex value of the hash sum. */
  public final String value;

  /**
   * The timestamp of the start of the hashing. This is to be used by file system consistency checks
   * to compare it to the lastModified timestamp of files.
   */
  public final long datetime;

  /** The hashing algorithm. */
  public final String algorithm;

  private Hash(String value, long datetime, String algorithm) {
    this.value = value.toLowerCase();
    this.datetime = datetime;
    this.algorithm = algorithm.toLowerCase();
  }

  /**
   * Create Hash.
   *
   * <p>Returns null, if value is null. If value != null the algorithm as well must not be null.
   *
   * @param value
   * @param datetime
   * @param algorithm
   * @return
   */
  public static Hash create(String value, long datetime, String algorithm) {
    if (value == null) {
      return null;
    }
    if (algorithm == null) {
      throw new IllegalArgumentException("Parameter algorithm must not be null.");
    }
    return new Hash(value, datetime, algorithm);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj instanceof Hash) {
      return ((Hash) obj).value.equals(this.value) && ((Hash) obj).algorithm.equals(this.algorithm);
    }
    return false;
  }

  @Override
  public int hashCode() {
    return value.hashCode();
  }

  public String getAlgorithm() {
    return algorithm;
  }
  ;

  @Override
  public String toString() {
    return getAlgorithm() + ":" + value;
  }
}
