/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 * Copyright (C) 2022 IndiScale GmbH <info@indiscale.com>
 * Copyright (C) 2022 Timm Fitschen <t.fitschen@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */
package org.caosdb.server.filesystem;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import org.caosdb.server.utils.Utils;

/**
 * Calculate hash sums for Blobs and Streams with various algorithms (e.g. SHA-512).
 *
 * @author Timm Fitschen <t.fitschen@indiscale.com>
 */
public class Hasher {

  /** ID of the SHA-512 algorithm. */
  public static final String Default = "SHA-512";

  /**
   * Factory for a Hasher which uses the SHA-512 algorithm.
   *
   * @return A SHA-512 Hasher.
   */
  public static final Hasher Default() {
    return create(Default);
  }

  /**
   * Hash the given file with the SHA-512 algorithm and return the {@link Hash}.
   *
   * @param file - file to be hashed
   * @return {@link Hash}
   * @throws IOException if the file reading failed
   */
  public static final Hash Default(File file) throws IOException {
    if (file == null || !file.exists() || !file.isFile()) {
      return null;
    }
    return create(Default).hash(file);
  }

  /**
   * Factory for a hasher with a particular algorithm.
   *
   * @param algorithm
   * @return {@link Hasher}
   */
  public static Hasher create(String algorithm) {
    try {
      final MessageDigest md = MessageDigest.getInstance(algorithm);
      return new Hasher(md);
    } catch (NoSuchAlgorithmException e) {
      throw new UnsupportedOperationException("Unknown algorithm: " + algorithm);
    }
  }

  private MessageDigest md;

  protected Hasher(MessageDigest md) {
    this.md = md;
  }

  /**
   * Hash the given byte stream with the algorithm of this hasher instance and return the {@link
   * Hash}.
   *
   * @param in - the byte stream to be hashed.
   * @return {@link Hash}
   * @throws IOException
   */
  public Hash hash(final InputStream in) throws IOException {
    return Hash.create(getHash(in), now(), md.getAlgorithm());
  }

  /**
   * Hash the given file with the algorithm of this hasher instance and return the {@link Hash}.
   *
   * @param file - the file to be hashed.
   * @return {@link Hash}
   * @throws IOException
   */
  public Hash hash(final File file) throws IOException {
    try (FileInputStream stream = new FileInputStream(file)) {
      return Hash.create(getHash(stream), now(), md.getAlgorithm());
    }
  }

  private String getHash(final InputStream stream) throws IOException {
    final byte[] buf = new byte[65536];
    int size = 0;
    while ((size = stream.read(buf)) != -1) {
      md.update(buf, 0, size);
    }
    final byte[] result = md.digest();
    final String ret = Utils.toHex(result);
    return ret;
  }

  protected long now() {
    return System.currentTimeMillis();
  }

  /**
   * Hash the given input stream with the algorithm of this hasher instance while writing the bytes
   * to the output stream and return the {@link Hash}.
   *
   * <p>This function is useful for combining the hashing and the transfer of a byte blob.
   *
   * @param inputStream - the byte stream to be hashed.
   * @param outputStream - copy the bytes into this output stream.
   * @return {@link Hash}
   * @throws IOException
   */
  public Hash hashAndWrite(InputStream inputStream, OutputStream outputStream) throws IOException {
    return Hash.create(getChecksumAndWrite(inputStream, outputStream), now(), md.getAlgorithm());
  }

  private String getChecksumAndWrite(InputStream inputStream, OutputStream outputStream)
      throws IOException {
    final byte[] buf = new byte[1024];
    int bufSize = 0;

    // write file and calculate hash
    while ((bufSize = inputStream.read(buf)) != -1) {
      md.update(buf, 0, bufSize);
      outputStream.write(buf, 0, bufSize);
    }

    // get hash result
    final byte[] result = md.digest();
    return Utils.toHex(result);
  }

  public void write(int data) {
    md.update((byte) data);
  }

  public void write(byte[] data) {
    md.update(data);
  }

  public void write(byte[] data, int pos, int pos2) {
    md.update(data, pos, pos2);
  }
}
