package org.caosdb.server.filesystem;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.file.Files;
import java.nio.file.LinkOption;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import org.apache.commons.io.FileUtils;
import org.caosdb.server.CaosDBException;
import org.caosdb.server.database.exceptions.TransactionException;
import org.caosdb.server.entity.Message;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.utils.ServerMessages;
import org.caosdb.server.utils.UndoHandler;
import org.caosdb.server.utils.Undoable;
import org.caosdb.server.utils.Utils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * A LOCAL file storage is an implementation which accesses the applications (i.e. the CaosDB
 * server's) host file system via the standard Java {@link File} and {@link Files} API.
 *
 * @author tf
 */
public abstract class LocalFileStorage implements FileStorageInterface {

  /**
   * Represents a file, directory or link in a local file storage.
   *
   * <p>Data is being cached forever (hence clients should request a new instance if the need more
   * up-to-date information).
   *
   * <p>The methods delegate most of the functionality back to the {@link LocalFileStorage} that
   * created them.
   *
   * @author Timm Fitschen <t.fitschen@indiscale.com>
   */
  static class LocalFSODescriptor implements RealFSODescriptorInterface {

    private final LocalFileStorage fileStorage;
    private File file = null;
    private ObjectType type = null;
    private final String key;
    private Hash hash = null;
    private Long size = null;
    private String name = null;
    private List<? extends RealFSODescriptorInterface> children = null;
    private String mimeType;

    public LocalFSODescriptor(LocalFileStorage fileStorage, String key) {
      this.fileStorage = fileStorage;
      this.key = key;
    }

    private LocalFSODescriptor(File file) {
      this.key = null;
      this.fileStorage = null;
      this.file = file;
    }

    private ObjectType _getObjectType() {
      if (type == null) {
        type = fileStorage.getObjectType(key);
      }
      return type;
    }

    @Override
    public File getFile() {
      if (this.file == null) {
        this.file = fileStorage.getFile(this.key);
      }
      return file;
    }

    @Override
    public ObjectType getObjectType() {
      if (type == null) {
        type = _getObjectType();
      }
      return type;
    }

    @Override
    public String getKey() {
      return key;
    }

    @Override
    public String getFileStorageId() {
      return fileStorage.getId();
    }

    @Override
    public Hash getHash() {
      if (hash == null) {
        hash = fileStorage.getHash(getKey());
      }
      return hash;
    }

    @Override
    public Long getSize() {
      if (size == null) {
        size = fileStorage.getSize(key);
      }
      return size;
    }

    @Override
    public String getName() {
      if (name == null) {
        name = fileStorage.getName(key);
      }
      return name;
    }

    @Override
    public List<? extends RealFSODescriptorInterface> listChildren() {
      if (children == null) {
        children = fileStorage.getChildren(key);
      }
      return children;
    }

    @Override
    public String getMimeType() {
      if (mimeType == null) {
        mimeType = fileStorage.getMimeType(key);
      }
      return mimeType;
    }

    @Override
    public Boolean exists() {
      return getFile().exists();
    }

    @Override
    public RealFSODescriptorInterface resolveLinkTarget() {
      try {
        Path resolvedPath = getFile().toPath().toRealPath();
        if (fileStorage.belongsToUs(resolvedPath)) {
          return new LocalFSODescriptor(fileStorage, fileStorage.getKey(resolvedPath));
        }
        for (String otherFileStorageId : FileSystem.getInstance().getFileStorages()) {
          if (!otherFileStorageId.equals(fileStorage.getId())) {
            FileStorageInterface otherFileStorage =
                FileSystem.getInstance().getFileStorage(otherFileStorageId);
            if (otherFileStorage instanceof LocalFileStorage) {
              if (((LocalFileStorage) otherFileStorage).belongsToUs(resolvedPath)) {
                // this file belongs to another file storage.
                return new LocalFSODescriptor(
                    (LocalFileStorage) otherFileStorage,
                    ((LocalFileStorage) otherFileStorage).getKey(resolvedPath));
              }
            }
          }
        }
        // We don't know this file.
        return new LocalFSODescriptor(resolvedPath.toFile());
      } catch (IOException e) {
        throw new RuntimeException(e);
      }
    }
  }

  private final Logger logger = LoggerFactory.getLogger(getClass());
  private final Path root;
  private final FileStorageCapabilities capabilities;
  private final String id;

  /**
   * @param id - this file storage's id.
   * @param capabilities - the capabilities of this particular file storage.
   * @param root - the root directory of this file storage in the host's file system.
   */
  public LocalFileStorage(String id, final FileStorageCapabilities capabilities, final Path root) {
    this.id = id;
    this.capabilities = capabilities;
    this.root = root.toAbsolutePath();
  }

  private boolean belongsToUs(Path path) {
    return path.startsWith(root);
  }

  private String getKey(Path path) {
    return root.relativize(path).toString();
  }

  public ObjectType getObjectType(String key) {
    File file = getFile(key);
    return getObjectType(file);
  }

  protected ObjectType getObjectType(File file) {
    return this.isSymlink(file)
        ? ObjectType.LINK
        : (file.isDirectory() ? ObjectType.DIRECTORY : ObjectType.FILE);
  }

  public String getMimeType(String key) {
    File file = getFile(key);
    switch (getObjectType(file)) {
      case DIRECTORY:
        return "inode/directory";
      case LINK:
        return "inode/link";
      default:
        return getMimeType(file);
    }
  }

  public List<? extends RealFSODescriptorInterface> getChildren(String key) {
    List<RealFSODescriptorInterface> children = new LinkedList<>();
    org.caosdb.server.filesystem.Path path = new org.caosdb.server.filesystem.Path(key);
    File directory = getFile(key);
    if (directory.isDirectory()) {
      for (String child : directory.list()) {
        RealFSODescriptorInterface childFso =
            new LocalFSODescriptor(this, path.append(child).toString());
        children.add(childFso);
      }
    }
    return children;
  }

  public String getName(String key) {
    return getFile(key).getName();
  }

  public Hash getHash(String key) {
    return getHash(key, Hasher.Default());
  }

  public Hash getHash(String key, Hasher hasher) {
    try {
      File file = getFile(key);
      if (getObjectType(file) == ObjectType.FILE) {
        return hasher.hash(file);
      }
    } catch (IOException e) {
      throw new RuntimeException("Could not calculate the files hash.", e);
    }
    return null;
  }

  protected File getFile(String key) {
    final Path path = getRoot().resolve(key);
    final File file = path.toFile();

    return file;
  }

  @Override
  public String getId() {
    return id;
  }

  private boolean isFree(final String fileKey) {
    return !getRoot().resolve(fileKey).toFile().exists();
  }

  @Override
  public FileStorageCapabilities getCapabilities() {
    return capabilities;
  }

  @Override
  public RealFSODescriptorInterface resolve(final String key) {
    if (key == null) {
      throw new NullPointerException("key was null");
    }
    return new LocalFSODescriptor(this, key);
  }

  @Override
  public Undoable move(final RealFSODescriptorInterface file, VirtualFSODescriptorInterface target)
      throws Message {
    String key = target.getKey();
    if (key == null) {
      key = target.getPath().toString();
      target.setKey(key);
    }

    final Path targetPath = getRoot().resolve(key);

    if (target.getObjectType() == ObjectType.DIRECTORY) {
      final File newDirectory = createDirs(targetPath.toFile());
      return new Undoable() {

        @Override
        public void undo() {
          // remove newly created directories
          if (newDirectory != null) {
            try {
              FileUtils.forceDelete(newDirectory);
            } catch (final Exception e) {
              logger.error(
                  "Undo was not successfull. Could not delete parent directory "
                      + newDirectory.getAbsolutePath(),
                  e);
            }
          }
        }

        @Override
        public void cleanUp() {}
      };
    } else if (target.getObjectType() == ObjectType.LINK) {
      final String newTargetPath = target.getLinkTargetPath().toString();
      final File newLink = targetPath.toFile();
      createSymlink(targetPath.toFile(), getRoot().resolve(newTargetPath).toFile());
      return new Undoable() {

        @Override
        public void undo() {
          // remove newly created directories
          if (newLink != null) {
            try {
              FileUtils.forceDelete(newLink);
            } catch (final Exception e) {
              logger.error(
                  "Undo was not successfull. Could not delete link " + newLink.getAbsolutePath(),
                  e);
            }
          }
        }

        @Override
        public void cleanUp() {}
      };

    } else {
      return rename(file.getFile(), targetPath.toFile());
    }
  }

  @Override
  public Undoable delete(final String fileKey) throws Message {
    final Path targetPath = getRoot().resolve(fileKey);

    return delete(targetPath.toFile(), true);
  }

  public Path getRoot() {
    return root;
  }

  protected File getTmpFile(final File target) {
    return root.resolve(".__tmp." + Utils.getUID() + target.getName()).toFile();
  }

  protected File createDirs(final File target) {
    File result = null;
    File dir = target;
    while (!dir.exists()) {
      result = dir;
      dir = dir.getParentFile();
    }
    target.mkdirs();
    return result;
  }

  protected Undoable rename(final File file, final File target) throws Message {
    final File backup;
    if (target.getAbsoluteFile().equals(file.getAbsoluteFile())) {
      return new UndoHandler() {
        @Override
        public void undo() {}

        @Override
        public void cleanUp() {}
      };
    }
    try {
      if (exists(target)) {
        // in case this is a update transaction, the old version of the file
        // must be stored somewhere until the transaction is done.
        final File tmp = getTmpFile(target);
        moveReplace(target, tmp);
        backup = tmp;
      } else {
        backup = null;
      }

      final File deleteParentDir = createDirs(target.getParentFile());

      moveReplace(file, target);
      return new UndoHandler() {

        protected File _backup = backup;

        @Override
        public void undo() {
          try {
            moveReplace(target, file);
            if (_backup != null) {
              moveReplace(_backup, target);
            }
          } catch (final Exception e) {
            if (_backup != null) {
              logger.error(
                  "Undo was not successful. Backup File in " + _backup.getAbsolutePath(), e);
              // prevent _backup from being deleted during cleanup
              _backup = null;
            } else {
              logger.error("Undo was not successfull.", e);
            }
          }

          // remove newly created directories
          if (deleteParentDir != null) {
            try {
              FileUtils.forceDelete(deleteParentDir);
            } catch (final Exception e) {
              logger.error(
                  "Undo was not successfull. Could not delete parent directory "
                      + deleteParentDir.getAbsolutePath(),
                  e);
            }
          }
        }

        @Override
        public void cleanUp() {
          if (_backup != null) {
            try {
              deleteFinally(_backup);
            } catch (final IOException e) {
              logger.error("Clean-up deletion was not successful", e);
            }
          }
        }

        @Override
        public String toString() {
          return "UndoHandler: restore backup";
        }
      };
    } catch (final Exception e) {
      logger.error("Error during moving a file to " + target.getAbsolutePath(), e);
      throw ServerMessages.CANNOT_MOVE_FILE_TO_TARGET_PATH;
    }
  }

  protected void moveReplace(final File file, final File target)
      throws IOException, InterruptedException {
    if (exists(target)) {
      FileUtils.forceDelete(target);
    }
    if (isDir(file)) {
      FileUtils.moveDirectory(file, target);
    } else if (isSymlink(file)) {
      moveSymlink(file, target);
    } else {
      FileUtils.moveFile(file, target);
    }
  }

  protected void moveSymlink(final File oldlink, final File newlink)
      throws IOException, InterruptedException {
    final File target = oldlink.toPath().toRealPath().toFile();
    createSymlink(newlink, target);
    callPosixUnlink(oldlink);
  }

  protected boolean exists(final File file) {
    return Files.exists(file.toPath(), LinkOption.NOFOLLOW_LINKS);
  }

  protected void deleteFinally(final File file) throws IOException {
    org.apache.commons.io.FileUtils.forceDelete(file);
  }

  protected void createSymlink(final File link, final File target) {
    if (target == null) {
      throw new NullPointerException("target was null.");
    }
    if (!target.exists()) {
      throw new TransactionException("target does not exist.");
    }
    try {

      final String targetPath = target.getCanonicalPath();
      final String linkPath = link.getCanonicalPath();
      final Process cmd =
          Runtime.getRuntime().exec(new String[] {"ln", "-s", targetPath, linkPath});

      if (cmd.waitFor() != 0) {
        if (!link.getParentFile().canWrite()) {
          throw new TransactionException(
              "symlink cannot be created. No permission to write in this directory: \n"
                  + link.getParent()
                  + "\nsymlink creation terminated with error.\n"
                  + Utils.InputStream2String(cmd.getErrorStream()));
        }
        throw new TransactionException(
            "symlink creation terminated with error.\n"
                + Utils.InputStream2String(cmd.getErrorStream()));
      }
    } catch (final InterruptedException e) {
      throw new TransactionException(e);
    } catch (final IOException e) {
      throw new TransactionException(e);
    }
  }

  protected void callPosixUnlink(final File file) throws IOException, InterruptedException {
    final Process cmd = Runtime.getRuntime().exec(new String[] {"unlink", file.getAbsolutePath()});
    if (cmd.waitFor() != 0) {
      logger.error("unlink return with non-zero exit code for " + file.getAbsolutePath());
      throw new CaosDBException("could not unlink " + file.getAbsolutePath());
    }
  }

  protected boolean isDir(final File file) {
    return Files.isDirectory(file.toPath(), LinkOption.NOFOLLOW_LINKS);
  }

  protected boolean isSymlink(final File file) {
    return Files.isSymbolicLink(file.toPath());
  }

  /**
   * Delete a file undoable. If file is a directory and 'recursive' is true the directory and all of
   * its sub folders are deleted. If file is a directory and 'recursive' is false file is deleted
   * unless it has sub folders. If file is an actual file (and not a directory) the file is deleted.
   * Otherwise, nothing happens.
   *
   * @param file
   * @param recursive
   * @return
   * @throws IOException
   * @throws InterruptedException
   * @throws CaosDBException
   */
  protected Undoable delete(final File file, final boolean recursive) throws Message {
    try {
      if (file.exists() && file.isFile()) {
        // delete file
        final File backup = getTmpFile(file);

        // move to tmp file
        moveReplace(file, backup);
        return new Undoable() {

          private File _backup = backup;

          @Override
          public void undo() {
            try {
              moveReplace(_backup, file);
            } catch (final Exception e) {
              logger.error(
                  "Undo was not successful. Backup File in " + _backup.getAbsolutePath(), e);
              // prevent _backup from being deleted during cleanup
              _backup = null;
            }
          }

          @Override
          public void cleanUp() {
            if (_backup != null) {
              try {
                deleteFinally(_backup);
              } catch (final IOException e) {
                logger.error("Clean-up deletion was not successful", e);
              }
            }
          }
        };
      } else if (file.exists() && file.isDirectory()) {
        // directory
        if (recursive && file.listFiles().length > 0) {
          // delete directory recursively with all of its sub folders
          final File backup = getTmpFile(file);

          // move to tmp folder
          moveReplace(file, backup);
          return new Undoable() {

            private File _backup = backup;

            @Override
            public void undo() {
              try {
                moveReplace(_backup, file);
              } catch (final Exception e) {
                logger.error(
                    "Undo was not successful. Backup File in " + _backup.getAbsolutePath(), e);
                // prevent _backup from being deleted during cleanup
                _backup = null;
              }
            }

            @Override
            public void cleanUp() {
              if (_backup != null) {
                try {
                  deleteFinally(_backup);
                } catch (final IOException e) {
                  logger.error("Clean-up deletion was not successful", e);
                }
              }
            }
          };
        } else if (file.listFiles().length == 0) {
          // non-recursive mode: delete empty directory
          deleteFinally(file);
          return new Undoable() {
            @Override
            public void undo() {
              file.mkdir();
            }

            @Override
            public void cleanUp() {}
          };
        }
      }
    } catch (final Exception e) {
      logger.error("Error during deleting the file " + file.getAbsolutePath(), e);
      throw ServerMessages.CANNOT_DELETE_FILE;
    }
    // do nothing
    return new Undoable() {

      @Override
      public void undo() {}

      @Override
      public void cleanUp() {}
    };
  }

  public Long getSize(String key) {
    File file = getFile(key);
    if (file == null) {
      throw new FileSystemException("Could not resolve key: " + key);
    }
    if (!exists(file)) {
      throw new FileSystemException("FSO does not exist: " + key);
    }
    if (file != null && !file.isDirectory()) {
      return file.length();
    }
    return null;
  }

  @Override
  public Long getSize(FSODescriptorInterface fso) {
    if (fso.getFileStorageId().equals(id)) {
      return getSize(fso.getKey());
    } else {
      throw new FileSystemException("Parameter 'fso' had wrong fileStorageId.");
    }
  }

  @Override
  public Hash getHash(FSODescriptorInterface fso, String algorithm) {
    if (fso.getFileStorageId().equals(id)) {
      return getHash(fso.getKey(), algorithm);
    } else {
      throw new FileSystemException("Parameter 'fso' had wrong fileStorageId.");
    }
  }

  private Hash getHash(String key, String algorithm) {
    File file = getFile(key);
    if (file == null) {
      throw new FileSystemException("Could not resolve key: " + key);
    }
    if (!exists(file)) {
      throw new FileSystemException("FSO does not exist: " + key);
    }
    if (!file.isDirectory()) {
      Hasher hasher = Hasher.create(algorithm);
      try {
        return hasher.hash(file);
      } catch (FileNotFoundException e) {
        throw new FileSystemException("FSO does not exist: " + key);
      } catch (IOException e) {
        throw new FileSystemException("Could not hash file: " + key, e);
      }
    }
    return null;
  }

  @Override
  public boolean exists(String key) {
    return !isFree(key);
  }

  /**
   * Guess the MIME type of a file using the unix program 'file -i'.
   *
   * @param file
   * @return
   */
  public String getMimeType(final File file) {
    if (!file.exists()) {
      throw new RuntimeException("File does not exist.");
    }
    try {
      final StringBuffer outputStringBuffer = new StringBuffer();
      final String path = file.getAbsolutePath();
      final Process cmd =
          Runtime.getRuntime().exec(new String[] {"file", "-b", "-i", "--dereference", path});
      final int status = cmd.waitFor();

      final Reader r = new InputStreamReader(cmd.getInputStream());
      final BufferedReader buf = new BufferedReader(r);
      String line;
      while ((line = buf.readLine()) != null) {
        outputStringBuffer.append(line);
      }
      if (outputStringBuffer.length() > 0) {
        final String ret = outputStringBuffer.toString().trim();
        if (ret.contains("cannot open")) {
          throw new RuntimeException("STATUS=" + status + " - " + ret);
        }
        return ret;
      } else {
        throw new RuntimeException("Output of `file` command was empty.");
      }

    } catch (final IOException e) {
      throw new RuntimeException(e);
    } catch (final InterruptedException e) {
      throw new RuntimeException(e);
    }
  }

  public static class LocalFSOIterator implements Iterator<LocalFSODescriptor> {

    private final File base;
    private File[] files = null;
    private LocalFSOIterator subfiles = null;
    private int i = 0;
    private final org.caosdb.server.filesystem.Path rootPath;
    private LocalFileStorage fileStorage;

    public LocalFSOIterator(
        LocalFileStorage fileStorage, File base, final org.caosdb.server.filesystem.Path rootPath) {
      this.fileStorage = fileStorage;
      this.rootPath = rootPath;
      this.base = base;
      this.files = this.base.listFiles();
      Arrays.sort(this.files);
    }

    @Override
    public boolean hasNext() {
      return this.files.length > this.i;
    }

    @Override
    public LocalFSODescriptor next() {
      if (this.subfiles != null) {
        if (this.subfiles.hasNext()) {
          return this.subfiles.next();
        } else {
          this.subfiles = null;
        }
      }
      final File ret = this.files[this.i++];
      if (ret.isDirectory()) {
        this.subfiles = new LocalFSOIterator(fileStorage, ret, this.rootPath.append(ret.getName()));
        if (this.subfiles.hasNext()) {
          return next();
        } else {
          return new LocalFSODescriptor(fileStorage, rootPath.append(ret.getName()).toString());
        }
      }
      return new LocalFSODescriptor(fileStorage, rootPath.append(ret.getName()).toString());
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException();
    }
  }

  @Override
  public Iterable<LocalFSODescriptor> list(String prefix) {
    final LocalFileStorage fileStorage = this;
    final File base = prefix == null ? this.root.toFile() : this.root.resolve(prefix).toFile();
    return new Iterable<LocalFSODescriptor>() {

      @Override
      public LocalFSOIterator iterator() {
        return new LocalFSOIterator(
            fileStorage, base, new org.caosdb.server.filesystem.Path(prefix == null ? "" : prefix));
      }
    };
  }
}
