package org.caosdb.server.filesystem;

import java.io.File;
import java.util.List;

/**
 * Represents an object in a file storage (i.e. a POSIX file system, an S3 bucket server, or similar
 * systems and services which can provide byte streams and file meta data) as it ACTUALLY IS, to the
 * best of our current knowledge.
 *
 * <p>The state of the files, directories and links are possibly changed by agents other than the
 * CaosDB server. That is why the information returned by this class's methods is always a snapshot
 * of the reality and it is impossible to guarantee the consistency between two subsequent method
 * calls. It is possible, that requesting the current size with {@link #getSize()} is already
 * out-dated when the {@link #getHash()} method is being called, because the files has been
 * rewritten in the mean time by another agent.
 *
 * <p>However, because it is generally not possible to guarantee the consistency, any data might be
 * cached by the implementation for some time.
 *
 * <p>An instance of this class is POSSIBLY a twin of an instance of {@link
 * VirtualFSODescriptorInterface} which represent the file system objects as the SHOULD BE (which is
 * being stored in the back-end). The file storages might know more/other file system objects than
 * the virtual file system of CaosDB. This is the case when not all files that are stored in a
 * particular file storage have been indexed by the CaosDB server (yet) or when files are being
 * deleted (by accident or on purpose).
 *
 * <p>N.B. it is possible that this class represents a non-existing file, e.g. before it is written
 * by the server, or because it has been deleted but the server knows a twin virtual file system
 * object.
 *
 * @author Timm Fitschen <t.fitschen@indiscale.com>
 */
public interface RealFSODescriptorInterface extends FSODescriptorInterface {

  // TODO
  public abstract File getFile();

  /**
   * Return true if the represented object exists, false if the represented object does not exist,
   * null if the existence cannot be determined.
   */
  public abstract Boolean exists();

  public abstract RealFSODescriptorInterface resolveLinkTarget();

  @Override
  public abstract List<? extends RealFSODescriptorInterface> listChildren();
}
