package org.caosdb.server.filesystem;

import java.io.File;
import java.nio.file.Path;
import java.util.regex.Pattern;
import org.caosdb.server.CaosDBServer;
import org.caosdb.server.ServerProperties;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.utils.Utils;

public class SharedFileStorage extends LocalFileStorage {

  private static final SharedFileStorage instance = new SharedFileStorage();
  public static final Pattern base32Pattern = Pattern.compile("^[-A-Z2-7]+$");

  public static SharedFileStorage getInstance() {
    return instance;
  }

  public SharedFileStorage() {
    super(
        "SHARED",
        FileStorageCapabilities.READ_ONLY,
        Path.of(CaosDBServer.getServerProperty(ServerProperties.KEY_SHARED_FOLDER)));
  }

  public File createSharedDirectory() {
    String dirName = Utils.getSecureFilename(15);
    Path root = getRoot();
    Path resultPath = root.resolve(dirName);
    File resultDir = resultPath.toFile();
    resultDir.mkdir();
    return resultDir;
  }

  /**
   * Get the file from the shared files folder.
   *
   * <p>Conditions under which null is returned:
   * <li>The file does not exist.
   * <li>The file is a folder.
   * <li>The requested path is just a file, without parent folders.
   * <li>The requested path is not normalized.
   * <li>The first component of the path does not match the base32 pattern for shared folders.
   *
   * @param path The path to the requested file.
   * @return File
   */
  public File getFromShared(final String path) {
    Path pathObj = (new File(path)).toPath();

    // Must have more than one component
    if (pathObj.getNameCount() < 2) {
      return null;
    }
    // Check for normalization
    if (!pathObj.equals(pathObj.normalize())) {
      return null;
    }
    // The first component of `path` must follow the Base32 pattern.
    String firstElement = pathObj.getName(0).toString();
    if (!base32Pattern.matcher(firstElement).matches()) {
      return null;
    }

    // known to this fs?
    if (!exists(path)) {
      return null;
    }

    // All safe, let's get the file already.
    RealFSODescriptorInterface ret = resolve(path);

    // Is it a regular file?
    if (ret.getObjectType() != ObjectType.FILE) {
      return null;
    }
    return ret.getFile();
  }
}
