package org.caosdb.server.filesystem;

import static org.apache.commons.io.FileUtils.deleteQuietly;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Path;
import org.apache.commons.fileupload.FileItemStream;
import org.apache.commons.io.FileExistsException;
import org.caosdb.server.CaosDBServer;
import org.caosdb.server.ServerProperties;
import org.caosdb.server.entity.Message;
import org.caosdb.server.entity.Message.MessageType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

class UploadFD extends TemporaryFSODescriptor {

  public UploadFD(
      final File tmpFile,
      final String fileId,
      final Hash hash,
      final Long size,
      final File tmpDir) {
    super(tmpFile, fileId, hash, size, tmpDir);
  }
}

class HashingFileOutputStream extends OutputStream {

  private Hasher hasher;
  private OutputStream outputStream;

  public HashingFileOutputStream(File file, Hasher hasher) throws FileNotFoundException {
    this.outputStream = new FileOutputStream(file);
    this.hasher = hasher;
  }

  @Override
  public void write(int data) throws IOException {
    outputStream.write(data);
    hasher.write(data);
  }

  @Override
  public void write(byte[] data) throws IOException {
    outputStream.write(data);
    hasher.write(data);
  }

  @Override
  public void write(byte[] data, int pos, int pos2) throws IOException {
    outputStream.write(data, pos, pos2);
    hasher.write(data, pos, pos2);
  }
}

public class TemporaryStorage {

  private final Logger logger = LoggerFactory.getLogger(getClass());
  private Path root;
  private static TemporaryStorage instance = new TemporaryStorage();

  public static class UploadFSODescriptorFactory {
    private final Logger logger = LoggerFactory.getLogger(getClass());
    private Path tmpDirPath;
    private Path tmpFilePath;
    private File tmpDir;
    private File tmpFile;
    private OutputStream stream;
    private String fileId;

    UploadFSODescriptorFactory(String prefix, String fileId) {
      this.tmpDirPath = TemporaryStorage.getInstance().getTmpDir(prefix);
      this.tmpFilePath = tmpDirPath.resolve(fileId);
      this.fileId = fileId;
    }

    public TemporaryFSODescriptor create() {
      File tmpFile = this.tmpFile;
      File tmpDir = this.tmpDir;
      this.tmpDir = null;
      this.tmpFile = null;
      cleanUp();
      return new UploadFD(tmpFile, fileId, null, tmpFile.length(), tmpDir);
    }

    public OutputStream getOutputStream() throws FileNotFoundException, FileExistsException {
      if (stream == null) {
        this.tmpDir = tmpDirPath.toFile();
        this.tmpDir.mkdirs();
        this.tmpFile = tmpFilePath.toFile();
        if (this.tmpFile.exists()) {
          throw new FileExistsException(tmpFile);
        }
        this.stream = new HashingFileOutputStream(tmpFile, Hasher.Default());
      }
      return stream;
    }

    public void cleanUp() {
      if (stream != null) {
        try {
          stream.close();
        } catch (IOException e) {
          logger.warn("Could not close the FileOutputStream", e);
        }
        stream = null;
      }

      if (tmpFile != null) {
        deleteQuietly(tmpFile);
        tmpFile = null;
      }

      if (tmpDir != null && tmpDir.exists() && tmpDir.list().length == 0) {
        deleteQuietly(tmpDir);
        tmpDir = null;
      }
    }
  }

  public static TemporaryStorage getInstance() {
    return instance;
  }

  public TemporaryStorage() {
    this.root = Path.of(CaosDBServer.getServerProperty(ServerProperties.KEY_TMP_FILES));
  }

  private Path getTmpDir(final String srid) {
    return getRoot().resolve(srid);
  }

  private Path getRoot() {
    return root;
  }

  public UploadFSODescriptorFactory createUploadFSODescriptorFactory(String prefix, String fileId) {
    return new UploadFSODescriptorFactory(prefix, fileId);
  }

  public TemporaryFSODescriptor upload(final FileItemStream item, final String srid)
      throws Message {
    final Path tmpDir = getTmpDir(srid);
    String tmpIdentifier = item.getName();
    final File tmpFile = tmpDir.resolve(tmpIdentifier).toFile();
    Long size = null;
    Hash hash = null;

    tmpFile.getParentFile().mkdirs();

    if (!tmpFile.isDirectory()) {
      final Hasher hasher = Hasher.Default();

      try (OutputStream outputStream = new FileOutputStream(tmpFile)) {
        try (InputStream inputStream = item.openStream()) {
          hash = hasher.hashAndWrite(inputStream, outputStream);
        }

      } catch (final IOException e) {
        logger.error("Error during the file upload.", e);
        throw new Message(MessageType.Error, "Error during the file upload.");
      }
      size = tmpFile.length();
    }

    return new UploadFD(tmpFile, tmpIdentifier, hash, size, tmpDir.toFile());
  }

  //
  //  public TemporaryFSODescriptor link(final File target, final String srid) {
  //    final Path tmpDir = getTmpDir(srid);
  //    final File tmpFile = tmpDir.resolve(Utils.getUID() + target.getName()).toFile();
  //    tmpDir.toFile().mkdirs();
  //
  //    createSymlink(tmpFile, target);
  //
  //    final Long size = tmpFile.length();
  //    Hash hash;
  //    try {
  //      hash = Hasher.Default(tmpFile);
  //    } catch (IOException e) {
  //      throw new TransactionException("Could not hash this file", e);
  //    }
  //
  //    return new LinkFD(tmpFile, hash, size, tmpDir.toFile());
  //  }

  public int getNumberOfTmpDirs() {
    String[] list = getRoot().toFile().list();
    if (list == null) {
      return 0;
    }
    return list.length;
  }
}
