package org.caosdb.server.filesystem.consistency;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import org.caosdb.datetime.UTCDateTime;
import org.caosdb.server.database.DatabaseAccessManager;
import org.caosdb.server.database.access.Access;
import org.caosdb.server.database.backend.transaction.GetFileIterator;
import org.caosdb.server.database.backend.transaction.GetVirtualFSO;
import org.caosdb.server.database.backend.transaction.SetFileCheckedTimestamp;
import org.caosdb.server.entity.EntityID;
import org.caosdb.server.filesystem.FSODescriptorInterface;
import org.caosdb.server.filesystem.FileStorageCapabilities;
import org.caosdb.server.filesystem.FileStorageInterface;
import org.caosdb.server.filesystem.FileSystem;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.filesystem.Hash;
import org.caosdb.server.filesystem.Path;
import org.caosdb.server.filesystem.RealFSODescriptorInterface;
import org.caosdb.server.filesystem.VirtualFSODescriptorInterface;
import org.caosdb.server.transaction.TransactionInterface;

abstract class ConsistencyCheckStrategy implements TransactionInterface, Runnable {

  private List<ConsistencyEventInterface> events = new LinkedList<>();
  private Access access;
  protected Exception exception;

  public abstract void execute();

  public void emit(ConsistencyEventInterface event) {
    events.add(event);
  }

  public List<ConsistencyEventInterface> getEvents() {
    return events;
  }

  void setAccess(Access access) {
    this.access = access;
  }

  Access getAccess() {
    return access;
  }

  @Override
  public void run() {
    execute();
  }

  public Exception getException() {
    return exception;
  }
}

class RunThroughInternalFileSystem extends ConsistencyCheckStrategy {

  private Long lastChecked;
  private Path path;
  private String fileStorageId;

  public RunThroughInternalFileSystem(
      Long lastChecked, Path path, FileStorageInterface fileStorage) {
    this.lastChecked = lastChecked == null ? System.currentTimeMillis() : lastChecked;
    this.path = path;
    this.fileStorageId = fileStorage == null ? null : fileStorage.getId();
  }

  @Override
  public void execute() {
    try {
      this.setAccess(DatabaseAccessManager.getInstance().acquireReadAccess(this));

      // test all files in file system.
      final GetFileIterator it = new GetFileIterator(fileStorageId, path, lastChecked);
      Iterator<VirtualFSODescriptorInterface> iterator =
          execute(it, this.getAccess()).getIterator();

      while (iterator != null && iterator.hasNext()) {
        long timestamp = System.currentTimeMillis();
        System.err.println("here");

        if (DatabaseAccessManager.whoHasReservedWriteAccess() != null) {
          // there is a thread waiting to write. pause this one and
          // try to acquire a new read access which will be granted when
          // the write thread is done.
          this.getAccess().release();
          this.setAccess(DatabaseAccessManager.getInstance().acquireReadAccess(this));
          iterator = execute(it, this.getAccess()).getIterator();
          System.err.println("here2");
        }

        final VirtualFSODescriptorInterface howItShouldBe = iterator.next();
        System.err.println("check: " + howItShouldBe.toString());

        check(howItShouldBe);

        execute(
            new SetFileCheckedTimestamp(howItShouldBe.getEntityId(), timestamp), this.getAccess());
        getAccess().commit();
      }
    } catch (final Exception e) {
      this.exception = e;
    } finally {
      this.getAccess().release();
    }
  }

  private void check(VirtualFSODescriptorInterface howItShouldBe) {
    RealFSODescriptorInterface howItActuallyIs = resolve(howItShouldBe);
    Boolean exists = howItActuallyIs.exists();
    if (exists != null && !exists) {
      emit(ConsistencyEvent.missingFile(howItActuallyIs));
      return;
    }

    if (howItShouldBe.getObjectType() != howItActuallyIs.getObjectType()) {
      emit(ConsistencyEvent.wrongObjectType(howItShouldBe, howItActuallyIs.getObjectType()));
      return;
    }

    if (howItActuallyIs.getObjectType() == ObjectType.LINK) {
      checkLinkIntact(howItShouldBe, howItActuallyIs);
    } else if (howItActuallyIs.getObjectType() == ObjectType.FILE) {
      checkSize(howItShouldBe, howItActuallyIs);
      checkHash(howItShouldBe, howItActuallyIs);
    } else if (howItActuallyIs.getObjectType() == ObjectType.DIRECTORY) {
      // checkDirectoryContent(howItShouldBe, howItActuallyIs);
    }
  }

  private void checkDirectoryContent(
      VirtualFSODescriptorInterface howItShouldBe, RealFSODescriptorInterface howItActuallyIs) {
    List<? extends VirtualFSODescriptorInterface> childrenHowItShouldBe =
        howItShouldBe.listChildren();
    Collections.sort(childrenHowItShouldBe);
    List<? extends RealFSODescriptorInterface> childrenHowItActuallyIs =
        howItActuallyIs.listChildren();
    Collections.sort(childrenHowItActuallyIs);
    boolean mismatch = false;
    if (childrenHowItActuallyIs.size() == childrenHowItShouldBe.size()) {
      Iterator<? extends RealFSODescriptorInterface> iterator1 = childrenHowItActuallyIs.iterator();
      Iterator<? extends VirtualFSODescriptorInterface> iterator2 =
          childrenHowItShouldBe.iterator();
      while (iterator1.hasNext()) {
        if (iterator1.next().compareTo(iterator2.next()) != 0) {
          // damn, found a mismatching child
          mismatch = true;
          break;
        }
      }
      if (!mismatch) {
        return;
      }
    }

    // we already know, that there are problems. We just try and figure out the exact nature of the
    // problems.
    childrenHowItActuallyIs = new ArrayList<>(childrenHowItActuallyIs);
    childrenHowItShouldBe = new ArrayList<>(childrenHowItShouldBe);
    Iterator<? extends VirtualFSODescriptorInterface> iChildrenHowItShouldBe =
        childrenHowItShouldBe.iterator();
    while (iChildrenHowItShouldBe.hasNext()) {
      Iterator<? extends RealFSODescriptorInterface> iChildrenHowItActuallyIs =
          childrenHowItActuallyIs.iterator();
      VirtualFSODescriptorInterface childHowItShouldBe = iChildrenHowItShouldBe.next();
      while (iChildrenHowItActuallyIs.hasNext()) {
        RealFSODescriptorInterface childHowItActuallyIs = iChildrenHowItActuallyIs.next();
        if (childHowItActuallyIs.compareTo(childHowItShouldBe) == 0) {
          // we have a match. remove from both lists
          iChildrenHowItActuallyIs.remove();
          iChildrenHowItShouldBe.remove();
        }
      }
    }

    // everything that is left in the lists is either...
    for (RealFSODescriptorInterface unknownObject : childrenHowItActuallyIs) {
      // ... an unknown object
      emit(ConsistencyEvent.unknownChild(unknownObject));
    }
    for (VirtualFSODescriptorInterface missingObject : childrenHowItShouldBe) {
      // ... or a missing object
      emit(ConsistencyEvent.missingChild(missingObject));
    }
  }

  private void checkHash(
      FSODescriptorInterface howItShouldBe, FSODescriptorInterface howItActuallyIs) {
    Hash hash1 = howItActuallyIs.getHash();
    Hash hash2 = howItShouldBe.getHash();
    if (hash1 != null && hash2 != null && !hash1.equals(hash2)) {
      emit(ConsistencyEvent.hashMismatch(howItActuallyIs));
    }
  }

  private void checkSize(
      FSODescriptorInterface howItShouldBe, FSODescriptorInterface howItActuallyIs) {
    Long size = howItActuallyIs.getSize();
    if (size != null && !size.equals(howItShouldBe.getSize())) {
      // size == null means we don't know.
      emit(ConsistencyEvent.sizeMismatch(howItActuallyIs));
    }
  }

  private void checkLinkIntact(
      VirtualFSODescriptorInterface howItShouldBe, RealFSODescriptorInterface howItActuallyIs) {
    EntityID linkTargetId = howItShouldBe.getLinkTarget();

    RealFSODescriptorInterface linkTargetActuallyIs = howItActuallyIs.resolveLinkTarget();
    Boolean exists = linkTargetActuallyIs.exists();
    if (exists != null && !exists) {
      // if link is broken (target is not existing/reachable from the host file system of the server
      // - in case of network file servers, it might be intact from a different host with different
      // mounts)
      emit(ConsistencyEvent.brokenLink(howItActuallyIs));
    }

    FSODescriptorInterface linkTargetShouldBe = retrieveFSODescriptor(linkTargetId);
    if (!(Objects.equals(
            linkTargetShouldBe.getFileStorageId(), linkTargetActuallyIs.getFileStorageId())
        && Objects.equals(linkTargetShouldBe.getKey(), linkTargetActuallyIs.getKey()))) {
      emit(ConsistencyEvent.linkTargetChanged(howItShouldBe));
    }

    if (linkTargetActuallyIs.getFileStorageId() == null) {
      // target does not belong to the CaosDB file system
      emit(ConsistencyEvent.linkToAlienTargetFromOuterSpace(howItActuallyIs));
    } else if (linkTargetActuallyIs
        .getFileStorageId()
        .equals(linkTargetShouldBe.getFileStorageId())) {
      // link crosses file storage boundaries
      emit(ConsistencyEvent.linkToOtherFileStorage(howItActuallyIs));
    }
  }

  private FSODescriptorInterface retrieveFSODescriptor(EntityID linkTargetId) {
    // TODO Auto-generated method stub
    return null;
  }

  private RealFSODescriptorInterface resolve(FSODescriptorInterface fso) {
    RealFSODescriptorInterface result = null;
    result = FileSystem.getInstance().resolve(fso.getFileStorageId(), fso.getKey());
    Boolean exists = FileSystem.exists(fso);
    if (exists != null && !exists) {
      // exists == null means we cannot tell.
      emit(ConsistencyEvent.missingFile(fso));
    }
    return result;
  }

  @Override
  public UTCDateTime getTimestamp() {
    // TODO Auto-generated method stub
    return null;
  }
}

/**
 * This {@link ConsistencyCheckStrategy} can only detect unknown files in particular file storages.
 *
 * @author tf
 */
class FindUnknownInFileStorage extends ConsistencyCheckStrategy {

  private FileStorageInterface fileStorage;
  private String directoryKey;

  /**
   * Detect unknown files in the given file storage and optionally under the given directory.
   *
   * <p>If the file storage does not have the {@link FileStorageCapabilities#list} capability, the
   * execution of this strategy emits a MISSING_CAPABILITY event.
   *
   * <p>If the file storage does not have the {@link FileStorageCapabilities#directory} capability
   * and the directory key is not null, the execution of this strategy emits a MISSING_CAPABILITY
   * event.
   *
   * @param fileStorage
   * @param directoryKey
   */
  public FindUnknownInFileStorage(FileStorageInterface fileStorage, String directoryKey) {
    this.fileStorage = fileStorage;
    this.directoryKey = directoryKey;
  }

  @Override
  public void execute() {
    try {
      if (fileStorage != null) {
        check(fileStorage);
        return;
      }

      // else check all
      for (String fileStorageId : FileSystem.getInstance().getFileStorages()) {
        FileStorageInterface fileStorage = FileSystem.getInstance().getFileStorage(fileStorageId);
        check(fileStorage);
      }
    } catch (Exception e) {
      this.exception = e;
    }
  }

  private void check(FileStorageInterface fileStorage) throws InterruptedException {
    if (!hasNecessaryCapabilities(fileStorage)) {
      return;
    }

    Iterator<? extends RealFSODescriptorInterface> iterator =
        fileStorage.list(directoryKey).iterator();
    while (iterator.hasNext()) {
      RealFSODescriptorInterface next = iterator.next();
      if (!isKnown(next)) {
        emit(ConsistencyEvent.unknownFile(next));
      }
    }
  }

  private boolean isKnown(RealFSODescriptorInterface next) throws InterruptedException {
    this.setAccess(DatabaseAccessManager.getInstance().acquireReadAccess(this));
    try {
      return execute(new GetVirtualFSO(next.getFileStorageId(), next.getKey()), getAccess())
          .isKnown();
    } finally {
      this.getAccess().release();
    }
  }

  private boolean hasNecessaryCapabilities(FileStorageInterface fileStorage) {
    if (!fileStorage.getCapabilities().list) {
      emit(ConsistencyEvent.missingCapability("LIST", fileStorage));
      return false;
    }
    if (directoryKey != null && !fileStorage.getCapabilities().directory) {
      emit(ConsistencyEvent.missingCapability("DIRECTORY", fileStorage));
      return false;
    }
    return true;
  }

  @Override
  public UTCDateTime getTimestamp() {
    // TODO Auto-generated method stub
    return null;
  }
}

public class ConsistencyCheck extends Thread {

  private Path belowPath;
  private Exception exception;
  private List<ConsistencyEventInterface> events;

  public ConsistencyCheck(Path belowPath) {
    this.belowPath = belowPath;
  }

  @Override
  public void run() {
    RunThroughInternalFileSystem check1 = new RunThroughInternalFileSystem(null, belowPath, null);
    check1.run();
    exception = check1.getException();
    events = check1.getEvents();
    if (exception != null) {
      return;
    }

    FindUnknownInFileStorage check2 = new FindUnknownInFileStorage(null, null);
    check2.run();
    exception = check2.getException();
    events.addAll(check2.getEvents());
  }

  public Exception getException() {
    return exception;
  }

  public List<ConsistencyEventInterface> getEvents() {
    return events;
  }
}
