package org.caosdb.server.filesystem.consistency;

import java.util.UUID;
import org.caosdb.server.filesystem.FSODescriptorInterface;
import org.caosdb.server.filesystem.FileStorageInterface;
import org.caosdb.server.filesystem.FileSystem;

public class ConsistencyEvent implements ConsistencyEventInterface {

  public static final String UNKNOWN = "UNKNOWN";
  private String fileStorageId;
  private String key;
  private String type;
  private String uuid = null;
  private Long timestamp;

  ConsistencyEvent(String type, String fileStorageId, String key) {
    this.timestamp = System.currentTimeMillis();
    this.type = type;
    //    this.parent = parent;
    this.fileStorageId = fileStorageId;
    this.key = key;
  }

  @Override
  public String getFileStorageId() {
    return fileStorageId;
  }

  @Override
  public String getKey() {
    return key;
  }

  @Override
  public String getEventType() {
    return type;
  }

  //
  //  @Override
  //  public ConsistencyEvent getParent() {
  //    if (parent == null && type.contains(":")) {
  //      String[] components = type.split(":");
  //      components = Arrays.copyOfRange(components, 0, components.length - 2);
  //      parent = new AbstractConsistencyEvent(String.join(":", components), fileStorageId, key);
  //    }
  //    return parent;
  //  }

  @Override
  public boolean isSubtypeOf(String eventType) {
    if (!eventType.endsWith(":")) {
      return this.type.startsWith(eventType + ":");
    }
    return this.type.startsWith(eventType);
  }

  public static final ConsistencyEventInterface missingCapability(
      String capability, FileStorageInterface fileStorage) {
    return new ConsistencyEvent("NEED_CAPABILITY:" + capability, fileStorage.getId(), null);
  }

  public static ConsistencyEventInterface unknownFile(FSODescriptorInterface next) {
    return new ConsistencyEvent(UNKNOWN, next.getFileStorageId(), next.getKey());
  }

  public static ConsistencyEventInterface missingFile(FSODescriptorInterface next) {
    return new ConsistencyEvent("MISSING", next.getFileStorageId(), next.getKey());
  }

  public static ConsistencyEventInterface wrongObjectType(
      FSODescriptorInterface fso, FileSystem.ObjectType type) {
    return new ConsistencyEvent(
        "CHANGED:TYPE:TO_" + type.toString(), fso.getFileStorageId(), fso.getKey());
  }

  public static ConsistencyEventInterface brokenLink(FSODescriptorInterface fso) {
    return new ConsistencyEvent("LINK:BROKEN", fso.getFileStorageId(), fso.getKey());
  }

  public static ConsistencyEventInterface linkToOtherFileStorage(FSODescriptorInterface fso) {
    return new ConsistencyEvent(
        "LINK:TARGET:IN_OTHER_FILE_STORAGE", fso.getFileStorageId(), fso.getKey());
  }

  public static ConsistencyEventInterface linkToAlienTargetFromOuterSpace(
      FSODescriptorInterface fso) {
    return new ConsistencyEvent(
        "LINK:TARGET:NOT_IN_FILE_STORAGE", fso.getFileStorageId(), fso.getKey());
  }

  public static ConsistencyEventInterface sizeMismatch(FSODescriptorInterface fso) {
    return new ConsistencyEvent("CHANGED:SIZE", fso.getFileStorageId(), fso.getKey());
  }

  public static ConsistencyEventInterface hashMismatch(FSODescriptorInterface fso) {
    return new ConsistencyEvent("CHANGED:HASH", fso.getFileStorageId(), fso.getKey());
  }

  public static ConsistencyEventInterface linkTargetChanged(FSODescriptorInterface fso) {
    return new ConsistencyEvent("CHANGED:LINK:TARGET", fso.getFileStorageId(), fso.getKey());
  }

  @Override
  public String getUUID() {
    if (uuid == null) {
      uuid = UUID.randomUUID().toString();
    }
    return uuid;
  }

  @Override
  public Long getTimestamp() {
    return timestamp;
  }

  public static ConsistencyEventInterface missingChild(FSODescriptorInterface missingObject) {
    return missingFile(missingObject);
  }

  public static ConsistencyEventInterface unknownChild(FSODescriptorInterface unknownObject) {
    return unknownFile(unknownObject);
  }

  @Override
  public String toString() {
    StringBuilder result = new StringBuilder();
    result.append(type);
    if (fileStorageId != null) {
      result.append(" - ");
      result.append(fileStorageId);
      if (key != null) {
        result.append(":");
        result.append(key);
      }
    }
    return result.toString();
  }
}
