package org.caosdb.server.jobs.core;

import org.caosdb.server.database.backend.transaction.GetFileEntityByPath;
import org.caosdb.server.database.exceptions.EntityDoesNotExistException;
import org.caosdb.server.entity.EntityInterface;
import org.caosdb.server.entity.InsertEntity;
import org.caosdb.server.entity.MagicTypes;
import org.caosdb.server.entity.Role;
import org.caosdb.server.entity.UpdateEntity;
import org.caosdb.server.filesystem.FSODescriptor;
import org.caosdb.server.filesystem.Path;
import org.caosdb.server.filesystem.VirtualFSODescriptorInterface;
import org.caosdb.server.jobs.EntityFlagJob;
import org.caosdb.server.jobs.JobAnnotation;
import org.caosdb.server.jobs.TransactionStage;
import org.caosdb.server.permissions.EntityACL;
import org.caosdb.server.transaction.WriteTransactionInterface;

@JobAnnotation(
    flag = "autoCreateDirs",
    defaultValue = "true",
    loadAlways = true,
    stage = TransactionStage.PRE_CHECK)
public class AutoCreateDirs extends EntityFlagJob {

  @Override
  protected void job(final String value) {
    final boolean createDirs = "true".equalsIgnoreCase(value);

    if (createDirs
        && (getEntity() instanceof InsertEntity || getEntity() instanceof UpdateEntity)
        && getEntity().hasFSODescriptor()) {
      createDirs(getEntity().getFSODescriptor());
    }
  }

  private void createDirs(final VirtualFSODescriptorInterface fso) {
    VirtualFSODescriptorInterface nextFSO = fso;
    String fileStorageId = nextFSO.getFileStorageId();
    Path parentDir = nextFSO.getPath().getParent();
    while (parentDir != null) {
      final VirtualFSODescriptorInterface parentFSO = exists(parentDir);
      if (parentFSO != null) {
        // the parent does exist.
        fso.setParentDirectory(parentFSO.getEntityId());
        break;
      }
      nextFSO = createDir(fileStorageId, parentDir, nextFSO);
      parentDir = parentDir.getParent();
    }
    if (parentDir == null) { // root of file system
      nextFSO.setParentDirectory(MagicTypes.ROOT_DIRECTORY.getId());
    }
  }

  private VirtualFSODescriptorInterface exists(final Path dirPath) {
    for (final EntityInterface e : getContainer()) {
      if (e.getRole() == Role.Directory
          && (e instanceof InsertEntity || e instanceof UpdateEntity)
          && e.getFSODescriptor().getPath().equals(dirPath)) {
        // Found the directory in the container
        return e.getFSODescriptor();
      }
    }
    final GetFileEntityByPath t = new GetFileEntityByPath(dirPath);
    try {
      execute(t);
      return t.getEntity().getFSODescriptor();
    } catch (final EntityDoesNotExistException e) {
      return null;
    }
  }

  private VirtualFSODescriptorInterface createDir(
      final String fileStorageId, final Path dirPath, final VirtualFSODescriptorInterface child) {
    final String name = dirPath.getLastSegment();
    final VirtualFSODescriptorInterface newFD = FSODescriptor.createDir(fileStorageId, dirPath);

    child.setParentDirectory(newFD.getEntityId());

    final EntityInterface newDir = new InsertEntity(name, Role.Directory);
    newDir.setId(((WriteTransactionInterface) getTransaction()).generateId());
    newDir.setEntityACL(EntityACL.getOwnerACLFor(getUser()));
    newDir.setFSODescriptor(newFD);
    getContainer().add(newDir);

    return newFD;
  }
}
