/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.jobs.core;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.io.Reader;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.caosdb.api.entity.v1.MessageCode;
import org.caosdb.server.CaosDBServer;
import org.caosdb.server.database.exceptions.TransactionException;
import org.caosdb.server.entity.Message;
import org.caosdb.server.entity.Message.MessageType;
import org.caosdb.server.entity.xml.ToElementable;
import org.caosdb.server.filesystem.FSODescriptor;
import org.caosdb.server.filesystem.FileSystem;
import org.caosdb.server.filesystem.Path;
import org.caosdb.server.filesystem.RealFSODescriptorInterface;
import org.caosdb.server.filesystem.consistency.ConsistencyCheck;
import org.caosdb.server.filesystem.consistency.ConsistencyConfig;
import org.caosdb.server.filesystem.consistency.ConsistencyConfig.EventsLogConfig;
import org.caosdb.server.filesystem.consistency.ConsistencyConfig.LogLevel;
import org.caosdb.server.filesystem.consistency.ConsistencyEventInterface;
import org.caosdb.server.jobs.FlagJob;
import org.caosdb.server.jobs.JobAnnotation;
import org.caosdb.server.transaction.Retrieve;
import org.caosdb.server.transaction.Transaction;
import org.caosdb.server.utils.Observable;
import org.caosdb.server.utils.Observer;
import org.caosdb.server.utils.Undoable;
import org.jdom2.Element;

@JobAnnotation(
    flag = "fileStorageConsistency",
    loadOnDefault = false,
    transaction = Retrieve.class,
    description =
        "Invokes a consistency test of the file storage. The value can have any of:\n"
            + "  - a string with a timeout in ms (prefixed -t): `-t 120000`\n"
            + "  - a test case which can be run if the server is in debug mode (prefixed -c)\n"
            + "  - a filesystem location whose children are checked recursively,\n"
            + "    relative to the caosroot dilesystem location.\n")
public class CheckFileStorageConsistency extends FlagJob {

  public static final Pattern parseArgs =
      Pattern.compile("\\s*-t\\s*([0-9]+)\\s*|\\s*-c\\s*([^\\s]+)\\s*|\\s*([^\\s]+\\s*)");
  private int timeout = 1000 * 30; // milliseconds
  private Path location = null;

  public static enum TestCase {
    FILE_DOES_NOT_EXIST,
    FILE_WAS_MODIFIED,
    FILE_WAS_MODIFIED_2,
    UNKNOWN_FILE
  };

  @Override
  protected void job(final String value) {
    if (value != null) {
      final Matcher matcher = parseArgs.matcher((value));
      while (matcher.find()) {
        if (matcher.group(1) != null) {
          // group one is timeout (-t$ms)
          try {
            this.timeout = Integer.valueOf(matcher.group(1)) * 1000;
          } catch (final NumberFormatException e) {
          }
        }
        if (matcher.group(2) != null && CaosDBServer.isDebugMode()) {
          // group two is TestCase (-c$testcase)
          try {
            test(TestCase.valueOf(matcher.group(2)));
            continue;
          } catch (final IllegalArgumentException e) {
          } catch (final IOException | Message e) {
            throw new TransactionException(e);
          }
        }
        if (matcher.group(3) != null) {
          // group three is file system location
          this.location = new Path(matcher.group(3));
        }
      }
    }

    ConsistencyCheck test = new ConsistencyCheck(this.location);
    test.start();
    try {
      test.join(this.timeout);
    } catch (final InterruptedException e1) {
      throw new TransactionException(e1);
    }

    if (test.isAlive()) {
      // add info: not read yet
      getContainer()
          .addMessage(
              new Message(
                  "Test took too long. The results will be written to './ConsistencyTest.xml'"));
    } else {
      // add info/warning/error
      final Exception e = test.getException();
      if (e != null) {
        throw new TransactionException(e);
      }

      getContainer().addMessage(createMessages(test));
    }
  }

  private ToElementable createMessages(ConsistencyCheck test) {
    final List<ConsistencyEventInterface> events = test.getEvents();
    return new ToElementable() {

      @Override
      public void addToElement(Element ret) {
        EventsLogConfig c = ConsistencyConfig.getInstance().eventsLogConfig;
        if (events.size() == 0) {
          if (location != null) {
            ret.addContent(
                new Message("File system below " + location + "/ is consistent.").toElement());
          } else {
            ret.addContent(new Message("File system is consistent.").toElement());
          }
        } else {
          for (ConsistencyEventInterface event : events) {
            LogLevel level = c.get(event.getEventType());
            switch (level) {
              case ERROR:
                ret.addContent(
                    new Message(
                            MessageType.Error, MessageCode.MESSAGE_CODE_UNKNOWN, event.toString())
                        .toElement());
                break;
              case WARN:
                ret.addContent(
                    new Message(
                            MessageType.Warning, MessageCode.MESSAGE_CODE_UNKNOWN, event.toString())
                        .toElement());
                break;
              case INFO:
                ret.addContent(
                    new Message(
                            MessageType.Info, MessageCode.MESSAGE_CODE_UNKNOWN, event.toString())
                        .toElement());
                break;
              default:
                // ignore
                break;
            }
          }
        }
      }
    };
  }

  private void test(final TestCase testCase) throws IOException, Message {
    switch (testCase) {
      case FILE_DOES_NOT_EXIST:
        final RealFSODescriptorInterface f1 =
            FileSystem.getInstance().resolve("default", "debug/test_file_storage_consistency");
        if (f1 == null) {
          return;
        }
        FSODescriptor target = new FSODescriptor(new Path("debug/somewhere.else"));
        Undoable undo1 = FileSystem.getInstance().getFileStorage(f1).move(f1, target);

        getTransaction()
            .acceptObserver(
                new Observer() {
                  @Override
                  public boolean notifyObserver(final String e, final Observable sender) {
                    if (e.equals(Transaction.CLEAN_UP)) {
                      undo1.undo();
                      return false;
                    }
                    return true;
                  }
                });
        break;

      case UNKNOWN_FILE:
        final File dir =
            FileSystem.getInstance().getFileStorage("default").resolve("test_unknown").getFile();

        dir.mkdirs();

        Undoable undo3 =
            new Undoable() {

              @Override
              public void undo() {
                try {
                  Undoable delete =
                      FileSystem.getInstance().getFileStorage("default").delete("test_unknown");
                  delete.cleanUp();
                } catch (Message e) {
                  e.printStackTrace();
                }
              }

              @Override
              public void cleanUp() {}
            };

        getTransaction()
            .acceptObserver(
                new Observer() {
                  @Override
                  public boolean notifyObserver(final String e, final Observable sender) {
                    if (e.equals(Transaction.CLEAN_UP)) {
                      undo3.undo();
                      return false;
                    }
                    return true;
                  }
                });
        break;
      case FILE_WAS_MODIFIED:
        final RealFSODescriptorInterface f2 =
            FileSystem.getInstance().resolve("default", "debug/test_file_storage_consistency");
        if (f2 == null) {
          return;
        }

        Reader r = new FileReader(f2.getFile());
        BufferedReader b = new BufferedReader(r);
        String oldContent = b.readLine();

        final File overridden = f2.getFile();
        Undoable undo2 =
            new Undoable() {

              @Override
              public void undo() {
                PrintStream out;
                try {
                  out = new PrintStream(overridden);
                  out.println(oldContent);
                  out.flush();
                  out.close();
                } catch (FileNotFoundException e) {
                  throw new RuntimeException(e);
                }
              }

              @Override
              public void cleanUp() {}
            };

        getTransaction()
            .acceptObserver(
                new Observer() {
                  @Override
                  public boolean notifyObserver(final String e, final Observable sender) {
                    if (e.equals(Transaction.CLEAN_UP)) {
                      undo2.undo();
                      return false;
                    }
                    return true;
                  }
                });

        final PrintStream out = new PrintStream(overridden);
        out.println("asdfasdfasdfasdfasdf");
        out.flush();
        out.close();

        break;
      case FILE_WAS_MODIFIED_2:
        final RealFSODescriptorInterface f3 =
            FileSystem.getInstance().resolve("default", "debug/test_file_storage_consistency");
        if (f3 == null) {
          return;
        }

        final File overridden2 = f3.getFile();

        try (final PrintStream out2 = new PrintStream(overridden2)) {
          out2.println("asdfasdfasdfasdfasdf");
        }
        break;

      default:
        break;
    }
  }
}
