/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.jobs.core;

import org.caosdb.server.entity.Message;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.filesystem.RealFSODescriptorInterface;
import org.caosdb.server.filesystem.VirtualFSODescriptorInterface;
import org.caosdb.server.jobs.FilesJob;
import org.caosdb.server.utils.EntityStatus;
import org.caosdb.server.utils.ServerMessages;

/**
 * Match the checksum and the size of a file record to the uploaded file's checksum and size.
 *
 * @author tf
 */
public class MatchFileProp extends FilesJob {
  @Override
  public final void run() throws Message {

    // does the entity have a file?
    if (getEntity().hasFSODescriptor()) {
      // normativeFileProperties is how the file must look like (path,
      // checksum, size)
      final VirtualFSODescriptorInterface normativeFileProperties = getEntity().getFSODescriptor();

      // descriptiveFileProperties is how the copied/uploaded/... file
      // ACTUALLY looks like. (path, checksum, size)
      final RealFSODescriptorInterface descriptiveFileProperties =
          getRealFSODescriptor(normativeFileProperties);

      if (descriptiveFileProperties != null) {
        checkChecksum(normativeFileProperties, descriptiveFileProperties);
        checkSize(normativeFileProperties, descriptiveFileProperties);
      }
    }
  }

  private void checkSize(
      final VirtualFSODescriptorInterface normFP, final RealFSODescriptorInterface descFP) {
    if (normFP.hasSize()) {
      // normFP and descFP have a size -> must be equal
      if (!descFP.getSize().equals(normFP.getSize())) {
        getEntity().setEntityStatus(EntityStatus.UNQUALIFIED);
        getEntity().addError(ServerMessages.SIZE_TEST_FAILED);
      }
    } else {
      normFP.setSize(descFP.getSize());
    }
  }

  private void checkChecksum(
      final VirtualFSODescriptorInterface normFP, final RealFSODescriptorInterface descFP) {
    if (descFP.getObjectType() == ObjectType.DIRECTORY) {
      return;
    }
    if (normFP.hasHash()) {
      // normFP's and descFP's checksum must be equal
      if (!descFP.getHash().equals(normFP.getHash())) {
        getEntity().addError(ServerMessages.CHECKSUM_TEST_FAILED);
      } else {
        normFP.updateHash(descFP.getHash());
      }
    } else if (descFP.getHash() != null) {
      // normFP doesn't have a checksum but descFP has -> copy to normFP
      normFP.setHash(descFP.getHash());
    }
  }
}
