/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.resource;

import static java.net.URLDecoder.decode;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Map;
import org.caosdb.server.database.backend.implementation.MySQL.ConnectionException;
import org.caosdb.server.database.exceptions.EntityDoesNotExistException;
import org.caosdb.server.database.misc.TransactionBenchmark;
import org.caosdb.server.entity.EntityID;
import org.caosdb.server.entity.EntityInterface;
import org.caosdb.server.entity.Message;
import org.caosdb.server.entity.Message.MessageType;
import org.caosdb.server.entity.RetrieveEntity;
import org.caosdb.server.entity.container.RetrieveContainer;
import org.caosdb.server.filesystem.FileSystem;
import org.caosdb.server.filesystem.FileSystem.ObjectType;
import org.caosdb.server.filesystem.Path;
import org.caosdb.server.filesystem.RealFSODescriptorInterface;
import org.caosdb.server.filesystem.VirtualFSODescriptorInterface;
import org.caosdb.server.permissions.EntityPermission;
import org.caosdb.server.transaction.Retrieve;
import org.caosdb.server.transaction.RetrieveSparseEntityByPath;
import org.caosdb.server.utils.ServerMessages;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.restlet.data.Disposition;
import org.restlet.data.MediaType;
import org.restlet.data.Status;
import org.restlet.representation.FileRepresentation;
import org.restlet.representation.Representation;

/**
 * Download files via GET method from the file system directly without making the detour through the
 * database. Other methods are not supported
 *
 * @author Timm Fitschen
 */
public class FileSystemResource extends AbstractCaosDBServerResource {

  public static Message ORPHANED_FILE_WARNING =
      new Message(
          MessageType.Warning,
          "Orphaned file. The file is not tracked. This is probably a harmless inconsistency but it might be a sign of other problems.");

  /**
   * Download a File from the CaosDBFileSystem. Only one File per Request.
   *
   * @author Timm Fitschen
   * @return InputRepresentation
   * @throws IOException
   */
  @Override
  protected final Representation httpGetInChildClass() throws Exception {
    try {
      final long t1 = System.currentTimeMillis();
      final Path path = getPath();

      final Element rootElem = generateRootElement();
      final Document doc = new Document(rootElem);
      EntityInterface entity = null;
      VirtualFSODescriptorInterface file = null;

      entity = getEntityByPath(path);
      if (entity.getFSODescriptor().getObjectType() == ObjectType.LINK) {
        entity = getEntityById(entity.getFSODescriptor().getLinkTarget());
      }
      file = entity.getFSODescriptor();

      if (file.getObjectType() == ObjectType.DIRECTORY) {
        String url = getUtils().getServerRootURI() + "/FileSystem/" + file.getPath().toString();
        if (!url.endsWith(Path.DEFAULT_PATH_SEPARATOR)) {
          url = url + Path.DEFAULT_PATH_SEPARATOR;
        }
        final Element folder = new Element("dir");
        folder.setAttribute("path", Path.DEFAULT_PATH_SEPARATOR + file.getPath().toString());
        folder.setAttribute("name", file.getPath().isEmpty() ? "/" : file.getName());
        folder.setAttribute("url", url);

        for (final VirtualFSODescriptorInterface child : file.listChildren()) {
          Element celem = null;
          switch (child.getObjectType()) {
            case DIRECTORY:
              celem = new Element("dir");
              celem.setAttribute("name", child.getName());
              celem.setAttribute("url", url + child.getName() + Path.DEFAULT_PATH_SEPARATOR);
              break;
            case LINK:
              celem = new Element("link");
              celem.setAttribute("name", child.getName());
              celem.setAttribute("url", url + child.getName());
              break;
            case FILE:
              celem = new Element("file");
              celem.setAttribute("name", child.getName());
              celem.setAttribute("url", url + child.getName());
              break;
            default:
              throw new UnsupportedOperationException(
                  "Unknown FileSystem.ObjectType: " + child.getObjectType().toString());
          }
          if (child.getEntityId().toString() != null) {
            celem.setAttribute("id", child.getEntityId().toString());
          }

          folder.addContent(celem);
        }
        rootElem.addContent(folder);

        final long t2 = System.currentTimeMillis();
        getBenchmark().addMeasurement(this.getClass().getSimpleName(), t2 - t1);
        rootElem.addContent(getBenchmark().toElement());
        return ok(doc);

      } else {
        if (entity.hasPermission(EntityPermission.RETRIEVE_FILE)) {
          // if (file.isReadable()) {
          // TODO check if FileDescriptor is READABLE
          final MediaType mt = MediaType.valueOf(file.getMimeType());
          RealFSODescriptorInterface fso =
              FileSystem.getInstance().resolve(file.getFileStorageId(), file.getKey());
          final FileRepresentation ret = new FileRepresentation(fso.getFile(), mt);
          ret.setDisposition(new Disposition(Disposition.TYPE_ATTACHMENT));

          return ret;
          //          } else {
          //            rootElem.addContent(toElement(file));
          //            return ok(doc);
          //          }
        } else {
          return error(ServerMessages.AUTHORIZATION_ERROR, Status.CLIENT_ERROR_FORBIDDEN);
        }
      }

    } catch (final EntityDoesNotExistException e) {
      return error(ServerMessages.ENTITY_DOES_NOT_EXIST, Status.CLIENT_ERROR_NOT_FOUND);
    }
  }

  private Path getPath() throws UnsupportedEncodingException {
    return new Path(
        decode(
            getRequest().getAttributes().containsKey("path")
                ? (String) getRequest().getAttributes().get("path")
                : "",
            "UTF-8"));
  }

  private String getReferenceString(VirtualFSODescriptorInterface file)
      throws UnsupportedEncodingException {
    return getReference().toString().replaceFirst(getPath().toString(), file.getPath().toString());
  }

  private Element toElement(final VirtualFSODescriptorInterface file) {
    // TODO
    return new Element("FileDescriptor");
  }

  TransactionBenchmark getBenchmark() {
    return TransactionBenchmark.getRootInstance().getBenchmark(getClass());
  }

  private EntityInterface getEntityById(EntityID id) throws Exception {
    final long t1 = System.currentTimeMillis();
    Map<String, String> flags = getFlags();
    flags.put("listDirectory", "true");
    RetrieveContainer c = new RetrieveContainer(getUser(), getTimestamp(), getSRID(), flags);
    c.add(new RetrieveEntity(id));
    Retrieve t = new Retrieve(c);
    t.execute();
    final long t2 = System.currentTimeMillis();
    getBenchmark().addMeasurement(this.getClass().getSimpleName() + ".getEntityById", t2 - t1);
    return t.getContainer().get(0);
  }

  /**
   * Throws EntityDoesNotExistException when there is not entity with that path.
   *
   * @param path
   * @return
   * @throws Exception
   */
  private EntityInterface getEntityByPath(final Path path) throws Exception {
    final long t1 = System.currentTimeMillis();
    final RetrieveSparseEntityByPath t = new RetrieveSparseEntityByPath(path);
    t.execute();
    final long t2 = System.currentTimeMillis();
    getBenchmark().addMeasurement(this.getClass().getSimpleName() + ".getEntityByPath", t2 - t1);
    return t.getEntity();
  }

  @Override
  protected Representation httpPostInChildClass(final Representation entity)
      throws ConnectionException, JDOMException {
    this.setStatus(org.restlet.data.Status.CLIENT_ERROR_METHOD_NOT_ALLOWED);
    return null;
  }
}
