/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.transaction;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import org.caosdb.server.CaosDBException;
import org.caosdb.server.database.DatabaseAccessManager;
import org.caosdb.server.database.access.Access;
import org.caosdb.server.database.backend.transaction.GetUpdateableChecksums;
import org.caosdb.server.database.backend.transaction.RetrieveSparseEntity;
import org.caosdb.server.database.backend.transaction.SetFileChecksum;
import org.caosdb.server.entity.EntityID;
import org.caosdb.server.entity.EntityInterface;
import org.caosdb.server.entity.Message;
import org.caosdb.server.entity.container.WritableContainer;
import org.caosdb.server.filesystem.FileSystem;
import org.caosdb.server.filesystem.Hash;
import org.caosdb.server.filesystem.VirtualFSODescriptorInterface;

/**
 * Retrieves all file without a checksum, calculates one and stores it to the database (method
 * 'run'). This is meant for asynchronous checksum calculation.
 *
 * @author tf
 */
public class ChecksumUpdater extends WriteTransaction
    implements Runnable, WriteTransactionInterface {

  private Boolean running = false;

  private static final ChecksumUpdater instance = new ChecksumUpdater();

  private ChecksumUpdater() {
    super(new WritableContainer());
  }

  /** Retrieves all file without a checksum, calculates one and stores it to the database */
  private void doRun() throws Exception {
    // this entity is updated at the end of each while loop
    EntityInterface fileEntity = getNextUpdateableFileEntity();
    while (true) {
      if (fileEntity == null) {
        // there is no next fileEntity
        return;
      }

      if (!isStillToBeUpdated(fileEntity)) {
        // entity for the next round in the while loop
        fileEntity = getNextUpdateableFileEntity();
        continue;
      }

      // we calculate the hash, even though we don't have reserved write access...
      final Hash checksum =
          FileSystem.getInstance()
              .resolve(
                  fileEntity.getFSODescriptor().getFileStorageId(),
                  fileEntity.getFSODescriptor().getKey())
              .getHash();
      if (checksum == null) {
        // TODO file system does not support hashing? Mark this file appropriately
        continue;
      }
      Access strongAccess = null;

      try {
        strongAccess = DatabaseAccessManager.getInstance().reserveWriteAccess(this);
        if (wasUpdatedInTheMeantime(fileEntity, strongAccess)) {
          // ... we calculated the hash too early :( try again
          continue;
        }

        fileEntity.getFSODescriptor().setHash(checksum);
        DatabaseAccessManager.getInstance().acquireWriteAccess(this);

        // update
        execute(new SetFileChecksum(fileEntity), strongAccess);
        strongAccess.commit();

      } finally {
        if (strongAccess != null) {
          strongAccess.release();
        }
      }

      // entity for the next round in the while loop
      fileEntity = getNextUpdateableFileEntity();
    }
  }

  @Override
  public void run() {
    try {
      synchronized (instance.running) {
        instance.running = true;
      }
      doRun();
    } catch (final Throwable t) {
      t.printStackTrace();
    } finally {
      synchronized (instance.running) {
        instance.running = false;
      }
    }
  }

  private EntityInterface getNextUpdateableFileEntity() throws InterruptedException {
    final Access weakAccess = DatabaseAccessManager.getInstance().acquireReadAccess(this);
    try {
      synchronized (instance.running) {

        // test for updatable checksums
        final EntityID id = execute(new GetUpdateableChecksums(), weakAccess).getID();
        if (id == null) {
          // nothing to be updated...
          instance.running = false;
          return null;
        }
        return execute(new RetrieveSparseEntity(id, null), weakAccess).getEntity();
      }
    } finally {
      weakAccess.release();
    }
  }

  private boolean wasUpdatedInTheMeantime(final EntityInterface fileEntity, final Access access)
      throws NoSuchAlgorithmException, IOException, CaosDBException, Message {
    final EntityInterface checkEntity =
        execute(new RetrieveSparseEntity(fileEntity), access).getEntity();
    final VirtualFSODescriptorInterface thatFP = checkEntity.getFSODescriptor();
    final VirtualFSODescriptorInterface thisFP = fileEntity.getFSODescriptor();

    return !(isStillToBeUpdated(checkEntity) && thatFP.getPath().equals(thisFP.getPath()));
  }

  private static boolean isStillToBeUpdated(final EntityInterface fileEntity) {
    return fileEntity.hasFSODescriptor() && !fileEntity.getFSODescriptor().hasHash();
  }

  public static void start() {
    synchronized (instance.running) {
      if (instance.running) {
        return;
      }
      instance.running = true;

      // run
      new Thread(instance).start();
    }
  }
}
