/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.utils;

import java.io.IOException;
import java.sql.SQLException;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.caosdb.datetime.UTCDateTime;
import org.caosdb.server.CaosDBServer;
import org.caosdb.server.database.DatabaseAccessManager;
import org.caosdb.server.database.access.Access;
import org.caosdb.server.database.backend.transaction.GetInfo;
import org.caosdb.server.database.backend.transaction.SyncStats;
import org.caosdb.server.database.exceptions.TransactionException;
import org.caosdb.server.database.proto.ProtoInfo;
import org.caosdb.server.entity.Message;
import org.caosdb.server.filesystem.TemporaryStorage;
import org.caosdb.server.transaction.TransactionInterface;
import org.jdom2.Element;

public class Info extends AbstractObservable implements Observer, TransactionInterface {

  private static Info instance = new Info();
  public static final String SYNC_DATABASE_EVENT = "SyncDatabaseEvent";
  private final Access access;
  public Logger logger = LogManager.getLogger(getClass());
  private UTCDateTime timestamp;

  @Override
  public boolean notifyObserver(final String e, final Observable o) {
    try {
      syncDatabase();
    } catch (final Exception exc) {
      logger.error(exc);
    }
    return true;
  }

  private Info() {
    this.timestamp = UTCDateTime.SystemMillisToUTCDateTime(System.currentTimeMillis());
    this.access = DatabaseAccessManager.getInfoAccess(this);
    try {
      syncDatabase();
    } catch (final Exception exc) {
      logger.error(exc);
    }
  }

  public static Info getInstance() {
    return instance;
  }

  private Long lastSync = 0L;
  private static int filesCount = -1;
  private static int propertiesCount = -1;
  private static int recordsCount = -1;
  private static int recordTypesCount = -1;
  private static int tmpFilesCount;
  private static String fssize = "";

  public static Integer getRecordsCount() throws Exception {
    return recordsCount;
  }

  public static final Integer getFilesCount() {
    return filesCount;
  }

  public static final Integer getPropertiesCount() {
    return propertiesCount;
  }

  public static final Integer getRecordTypesCount() {
    return recordTypesCount;
  }

  public long timeSinceSync() {
    long time = System.currentTimeMillis();
    return time - lastSync;
  }

  public void syncDatabase() throws Exception {
    synchronized (lastSync) {
      if (timeSinceSync() < 10000) {
        return;
      }
      lastSync = System.currentTimeMillis();
    }

    final ProtoInfo i = new ProtoInfo();
    final GetInfo t = new GetInfo(i);
    execute(t, access);
    filesCount = i.filesCount;
    recordsCount = i.recordsCount;
    recordTypesCount = i.recordTypesCount;
    propertiesCount = i.propertiesCount;
    fssize = Utils.getReadableByteSize(i.fssize);
    tmpFilesCount = countTmpFiles();

    notifyObservers(SYNC_DATABASE_EVENT);
  }

  private static int countTmpFiles() throws IOException {
    return TemporaryStorage.getInstance().getNumberOfTmpDirs();
  }

  public static Element toElement() throws Exception {
    return toElement(false);
  }

  /**
   * Generates an XML element showing the information for the /Info resource in CaosDB.
   *
   * @return An XML element containing:
   *     <ul>
   *       <li>The number of records
   *       <li>The number of properties
   *       <li>The number of record types
   *       <li>The number of files
   *       <li>The total file size
   *       <li>The number of temp files
   *     </ul>
   */
  public static Element toElement(final boolean tree) throws Exception, SQLException {
    if (filesCount == -1 && recordsCount == -1) {
      getInstance().syncDatabase();
    }
    final Element info = new Element("Stats");
    final Element counts = new Element("counts");
    counts.setAttribute("records", Integer.toString(recordsCount));
    counts.setAttribute("properties", Integer.toString(propertiesCount));
    counts.setAttribute("recordTypes", Integer.toString(recordTypesCount));
    counts.setAttribute("files", Integer.toString(filesCount));
    counts.setAttribute("fssize", fssize);
    counts.setAttribute("tmpfiles", Integer.toString(tmpFilesCount));
    if (CaosDBServer.isDebugMode()) {
      counts.setAttribute("debug", "true");
    }
    info.addContent(counts);
    return info;
  }

  public void syncDatabase(final ServerStat stat) throws TransactionException, Message {

    final Access access = getInstance().access;

    final SyncStats t = new SyncStats(stat);
    execute(t, access);
  }

  @Override
  public void execute() throws Exception {
    syncDatabase();
  }

  @Override
  public UTCDateTime getTimestamp() {
    return timestamp;
  }
}
