/*
 * ** header v3.0
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2018 Research Group Biomedical Physics,
 * Max-Planck-Institute for Dynamics and Self-Organization Göttingen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * ** end header
 */
package org.caosdb.server.utils;

import static java.net.URLDecoder.decode;
import static java.net.URLEncoder.encode;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.text.DecimalFormat;
import java.util.Scanner;
import java.util.regex.Pattern;
import org.apache.commons.codec.binary.Base32;
import org.caosdb.server.CaosDBServer;
import org.caosdb.server.ServerProperties;
import org.caosdb.server.entity.Message;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;

/** Utility functions. */
public class Utils {

  /** Secure random number generator, for secret random numbers. */
  private static final SecureRandom srand = new SecureRandom();

  /**
   * Regular expression pattern that checks a mail for RFC822 compliance.
   *
   * <p>Taken from: http://www.ex-parrot.com/~pdw/Mail-RFC822-Address.html
   *
   * <p>This belongs to the perl module "Mail::RFC822::Address". It is copyright 2001-2002 by Paul
   * Warren. The license can be found here:
   * http://www.ex-parrot.com/~pdw/Mail-RFC822-Address/Mail-RFC822-Address.html#copyright%20and%20license
   */
  public static final Pattern EMAIL_PATTERN =
      Pattern.compile(
          "(?:(?:\\r\\n)?[ \\t])*(?:(?:(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*))*@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*|(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)*\\<(?:(?:\\r\\n)?[ \\t])*(?:@(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*(?:,@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*)*:(?:(?:\\r\\n)?[ \\t])*)?(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*))*@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*\\>(?:(?:\\r\\n)?[ \\t])*)|(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)*:(?:(?:\\r\\n)?[ \\t])*(?:(?:(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*))*@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*|(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)*\\<(?:(?:\\r\\n)?[ \\t])*(?:@(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*(?:,@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*)*:(?:(?:\\r\\n)?[ \\t])*)?(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*))*@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*\\>(?:(?:\\r\\n)?[ \\t])*)(?:,\\s*(?:(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*))*@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*|(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)*\\<(?:(?:\\r\\n)?[ \\t])*(?:@(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*(?:,@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*)*:(?:(?:\\r\\n)?[ \\t])*)?(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\"(?:[^\\\"\\r\\\\]|\\\\.|(?:(?:\\r\\n)?[ \\t]))*\"(?:(?:\\r\\n)?[ \\t])*))*@(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*)(?:\\.(?:(?:\\r\\n)?[ \\t])*(?:[^()<>@,;:\\\\\".\\[\\] \\000-\\031]+(?:(?:(?:\\r\\n)?[ \\t])+|\\Z|(?=[\\[\"()<>@,;:\\\\\".\\[\\]]))|\\[([^\\[\\]\\r\\\\]|\\\\.)*\\](?:(?:\\r\\n)?[ \\t])*))*\\>(?:(?:\\r\\n)?[ \\t])*))*)?;\\s*)");

  /**
   * Check if email is RFC822 compliant.
   *
   * @param email
   * @return true when the email given by the String email is RFC822 compliant.
   */
  public static boolean isRFC822Compliant(final String email) {
    return EMAIL_PATTERN.matcher(email).matches();
  }

  /**
   * Converts an Element into a String. For debugging purposes.
   *
   * @param e The Element
   * @return String representation of the element.
   */
  public static String element2String(final Element e) {
    final Element par = e.getParentElement();
    e.detach();
    final Document d = new Document(e);
    try {
      return document2String(d);
    } finally {
      d.detachRootElement();
      if (par != null) {
        par.addContent(e);
      }
    }
  }

  /**
   * Converts a Document into a String. For debugging purposes.
   *
   * @param document
   * @return String representation of the document.
   */
  public static String document2String(final Document document) {
    if (document != null && document.hasRootElement()) {
      final XMLOutputter outputter = new XMLOutputter();
      final Format newFormat = Format.getPrettyFormat();
      outputter.setFormat(newFormat);
      return outputter.outputString(document);
    }
    return "";
  }

  /**
   * @param alg The hashing algorithm implementation.
   * @param string The string to be hashed.
   * @param salt Cryptographic salt.
   * @param it Number of iterations. Zero and negative integers don't make any sense. Therefore, if
   *     it<=1 the hashing is done once.
   * @return Hashed string, or null if string is null
   */
  public static String hash(
      final MessageDigest md, final String string, final String salt, final int it) {
    if (string == null) {
      return null;
    }
    md.update(string.getBytes());
    if (salt != null) {
      md.update(salt.getBytes());
    }
    byte[] result = md.digest();

    for (int i = 1; i < it; i++) {
      md.update(result);
      result = md.digest();
    }
    return toHex(result);
  }

  /**
   * This function returns the SHA512 hash of a given string, salt and number of iterations. This
   * function terminates the application completely with an error code when this hashing algorithm
   * is not available.
   *
   * @param string The string to be hashed.
   * @param salt Cryptographic salt.
   * @param it Number of iterations. Zero and negative integers don't make any sense. Therefore, if
   *     it<=1 the hashing is done once.
   * @return Hashed string, or null if string is null
   */
  public static String sha512(final String string, final String salt, final int it) {
    try {
      return hash(MessageDigest.getInstance("SHA-512"), string, salt, it);
    } catch (final NoSuchAlgorithmException e) {
      // This algorithm is a must have.
      e.printStackTrace();
      System.exit(1);
      return null;
    }
  }

  /**
   * Generate a UID (Hexadecimal).
   *
   * <p>Currently this generates two random long numbers and concatenates their hex representations.
   * This function is thread-safe.
   *
   * @return The UID as a String.
   */
  public static String getUID() {
    return Long.toHexString(srand.nextLong()) + Long.toHexString(srand.nextLong());
  }

  /**
   * Generate a secure filename (base32 letters and numbers).
   *
   * <p>Very similar to getUID, but uses cryptographic random number instead, also nicely formats
   * the resulting string.
   *
   * @param byteSize How many bytes of random bits shall be generated.
   * @return The filename as a String.
   */
  public static String getSecureFilename(int byteSize) {
    // get random bytes
    byte[] bytes = new byte[byteSize];
    srand.nextBytes(bytes);

    // Encode to nice letters
    // TODO use StringBuilder and iterate over filename.charArray
    String filename = (new Base32()).encodeToString(bytes);
    filename = filename.replaceAll("=+", "");

    // Split up into chunks of 8
    int chunksize = 8;
    int len = filename.length();
    String[] parts = new String[(int) Math.ceil(len * 1.0 / chunksize)];
    for (int i = 0; i < parts.length; ++i) {
      parts[i] = filename.substring(i * chunksize, Math.min(len, (i + 1) * chunksize));
    }
    filename = new String(parts[0]);
    for (int i = 1; i < parts.length; ++i) {
      filename += "-" + parts[i];
    }

    return filename;
  }

  /** Helper function calling InputStream2String for charset UTF-8. */
  public static String InputStream2String(final InputStream is) {
    return InputStream2String(is, "UTF-8");
  }

  /**
   * Return a string from the input stream is up to the next LF (line feed, \\A) character.
   *
   * @param is The input stream
   * @return The detected string or an empty string if the stream has become empty.
   */
  public static String InputStream2String(final InputStream is, String charset) {
    try (Scanner s = new Scanner(is, charset)) {
      s.useDelimiter("\\A");
      return s.hasNext() ? s.next() : "";
    }
  }

  /** Helper function to convert a String into an UTF-8 encoded ByteArrayInputStream. */
  public static InputStream String2InputStream(final String s) throws UnsupportedEncodingException {
    return new ByteArrayInputStream(s.getBytes("UTF-8"));
  }

  /**
   * Exception-free URI-encoding of strings using UTF-8. Any UnsupportedEncodingException will
   * terminate the server since this would by no means be an acceptable run-time environment if
   * UTF-8 is not available.
   *
   * @param string
   * @return encoded string.
   * @see {@link Utils#URLDecodeWithUTF8(String)}
   */
  public static String URLEncodeWithUTF8(final String string) {
    try {
      return encode(string, "UTF-8");
    } catch (final UnsupportedEncodingException e) {
      e.printStackTrace();
      System.exit(1);
      return null;
    }
  }

  /**
   * Exception-free decoding of URI-encoded Strings with UTF-8. Any UnsupportedEncodingException
   * will terminate the server since this would by no means be an acceptable run-time environment if
   * UTF-8 is not available.
   *
   * @param string
   * @return decoded string.
   * @see {@link Utils#URLEncodeWithUTF8(String)}
   */
  public static String URLDecodeWithUTF8(final String string) {
    try {
      return decode(string, "UTF-8");
    } catch (final UnsupportedEncodingException e) {
      e.printStackTrace();
      System.exit(1);
      return null;
    }
  }

  /** Intended to compute the maximum exponent for the required conversion. */
  public static double log(final int base, final Long x) {
    return Math.log(x) / Math.log(base);
  }

  /**
   * Converts file sizes to human readably sizes.
   *
   * <p>For example, getReadableByteSize(2048) == "2KiB".
   */
  public static String getReadableByteSize(final Long fssize) {
    if (fssize == null) {
      return null;
    } else if (fssize == 0) {
      return "0B";
    }

    final String[] s = new String[] {"B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"};
    final int i = (int) Math.floor(log(1024, fssize));
    String unit = "1024^" + i;
    if (i < s.length) {
      unit = s[i];
    } else {
      // leave the unit as 1024^x if there is no prefix available
    }

    final double ret = fssize / Math.pow(1024, i);
    return new StringBuilder()
        .append(new DecimalFormat("0.##").format(ret))
        .append(unit)
        .toString();
  }

  /**
   * Determines the strength of a password. Currently this just throws an error when the String
   * `password` does not at least contain - An uppercase char - A lowercase char - A number - A
   * punctuation char and if the length is not at least 8 characters.
   *
   * <p>correcthorsebatterystaple is also rejected.
   *
   * @param password The password to be checked.
   */
  public static void checkPasswordStrength(final String password) throws Message {
    String regex = CaosDBServer.getServerProperty(ServerProperties.KEY_PASSWORD_STRENGTH_REGEX);

    if (password.equals("correcthorsebatterystaple")) {
      throw ServerMessages.PASSWORD_TOO_WEAK("PWNED!");
    }

    if (!password.matches(regex)) {
      throw ServerMessages.PASSWORD_TOO_WEAK(
          CaosDBServer.getServerProperty(ServerProperties.KEY_PASSWORD_WEAK_MESSAGE));
    }
  }

  /**
   * Convert a byte array to HEX
   *
   * @param b
   * @return hexadecimal (as String).
   */
  public static String toHex(final byte[] binary) {
    final StringBuffer buf = new StringBuffer(2 * binary.length + 1);
    for (final byte b : binary) {
      int nibble = b >> 4 & 0x0F;
      if (nibble >= 0 && nibble < 10) {
        buf.append((char) ('0' + nibble));
      } else {
        buf.append((char) ('a' + (nibble - 10)));
      }
      nibble = b & 0x0F;
      if (nibble >= 0 && nibble < 10) {
        buf.append((char) ('0' + nibble));
      } else {
        buf.append((char) ('a' + (nibble - 10)));
      }
    }
    return buf.toString();
  }
}
